
## Copyright 2020    Informer (hhttps://github.com/zhouhaoyi/Informer2020/tree/main/models)
## Code modified for align the notation and the batch generation
## extended to all present in informer, autoformer folder
from torch import  nn
import torch

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from typing import List,Union

from .informer.encoder import Encoder, EncoderLayer, ConvLayer
from .informer.decoder import Decoder, DecoderLayer
from .informer.attn import FullAttention, ProbAttention, AttentionLayer
from .informer.embed import DataEmbedding
from ..data_structure.utils import beauty_string
#from .utils import Embedding_cat_variables not used here, custom cat embedding
from .utils import  get_scope

    
  
class Informer(Base):
    handle_multivariate = True
    handle_future_covariates = True
    handle_categorical_variables = True
    handle_quantile_loss = True
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    
    def __init__(self, 
                 d_model:int,
                 hidden_size:int,
                 n_layer_encoder:int,
                 n_layer_decoder:int,
                 mix:bool=True,
                 activation:str='torch.nn.ReLU',
                 remove_last = False,
                 attn: str='prob',
                 distil:bool=True,
                 factor:int=5,
                 n_head:int=1,
                 dropout_rate:float=0.1,
                
                 **kwargs)->None:
        """Initialize the model with specified parameters. hhttps://github.com/zhouhaoyi/Informer2020/tree/main/models
        
        Args:
            d_model (int): The dimensionality of the model.
            hidden_size (int): The size of the hidden layers.
            n_layer_encoder (int): The number of layers in the encoder.
            n_layer_decoder (int): The number of layers in the decoder.
            mix (bool, optional): Whether to use mixed attention. Defaults to True.
            activation (str, optional): The activation function to use. Defaults to 'torch.nn.ReLU'.
            remove_last (bool, optional): Whether to remove the last layer. Defaults to False.
            attn (str, optional): The type of attention mechanism to use. Defaults to 'prob'.
            distil (bool, optional): Whether to use distillation. Defaults to True.
            factor (int, optional): The factor for attention. Defaults to 5.
            n_head (int, optional): The number of attention heads. Defaults to 1.
            dropout_rate (float, optional): The dropout rate. Defaults to 0.1.
            **kwargs: Additional keyword arguments.
        
        Raises:
            ValueError: If any of the parameters are invalid.
        
        Notes:
            Ensure to set up split_params: shift: ${model_configs.future_steps} as it is required!!
        """
   
        super().__init__(**kwargs)
        self.save_hyperparameters(logger=False)
        beauty_string("BE SURE TO SETUP split_params:  shift:  ${model_configs.future_steps} BECAUSE IT IS REQUIRED",'info',True)
       
        self.remove_last = remove_last
        
        
        self.enc_embedding = DataEmbedding(self.past_channels, d_model, self.embs_past, dropout_rate)
        self.dec_embedding = DataEmbedding(self.future_channels, d_model, self.embs_fut, dropout_rate)
        # Attention
        Attn = ProbAttention if attn=='prob' else FullAttention
        # Encoder
        self.encoder = Encoder(
            [
                EncoderLayer(
                    AttentionLayer(Attn(False, factor, attention_dropout=dropout_rate, output_attention=False), 
                                d_model, n_head, mix=False),
                    d_model,
                    hidden_size,
                    dropout=dropout_rate,
                    activation=activation
                ) for _ in range(n_layer_encoder)
            ],
            [
                ConvLayer(
                    d_model
                ) for _ in range(n_layer_encoder-1)
            ] if distil else None,
            norm_layer=torch.nn.LayerNorm(d_model)
        )
        # Decoder
        self.decoder = Decoder(
            [
                DecoderLayer(
                    AttentionLayer(Attn(True, factor, attention_dropout=dropout_rate, output_attention=False), 
                                d_model, n_head, mix=mix),
                    AttentionLayer(FullAttention(False, factor, attention_dropout=dropout_rate, output_attention=False), 
                                d_model, n_head, mix=False),
                    d_model,
                    hidden_size,
                    dropout=dropout_rate,
                    activation=activation,
                )
                for _ in range(n_layer_decoder)
            ],
            norm_layer=torch.nn.LayerNorm(d_model)
        )

        self.projection = nn.Linear(d_model, self.out_channels*self.mul, bias=True)
        
                
        
    def can_be_compiled(self):
        return True  
        
    def forward(self,batch): 
        #x_enc, x_mark_enc, x_dec, x_mark_dec,enc_self_mask=None, dec_self_mask=None, dec_enc_mask=None):
        
        x_enc = batch['x_num_past'].to(self.device)
        idx_target_future = batch['idx_target_future'][0]

        if 'x_cat_past' in batch.keys():
            x_mark_enc = batch['x_cat_past'].to(self.device)
        else:
            x_mark_enc = None

        enc_self_mask = None
        
        x_dec = batch['x_num_future'].to(self.device)
        x_dec[:,-self.future_steps:,idx_target_future] = 0
        
        
        if 'x_cat_future' in batch.keys():
            x_mark_dec = batch['x_cat_future'].to(self.device)
        else:
            x_mark_dec = None
        dec_self_mask = None
        dec_enc_mask = None
        
        
        if self.remove_last:
            idx_target = batch['idx_target'][0]
            x_start = x_enc[:,-1,idx_target].unsqueeze(1)
            x_enc[:,:,idx_target]-=x_start   
        
    
        enc_out = self.enc_embedding(x_enc, x_mark_enc)
        enc_out, attns = self.encoder(enc_out, attn_mask=enc_self_mask)

        dec_out = self.dec_embedding(x_dec, x_mark_dec)
        dec_out = self.decoder(dec_out, enc_out, x_mask=dec_self_mask, cross_mask=dec_enc_mask)
        dec_out = self.projection(dec_out)
        
        # dec_out = self.end_conv1(dec_out)
        # dec_out = self.end_conv2(dec_out.transpose(2,1)).transpose(1,2)
        
        #import pdb
        #pdb.set_trace()
        res = dec_out[:,-self.future_steps:,:].unsqueeze(3)
        if self.remove_last:
            res+=x_start.unsqueeze(1)
        BS = res.shape[0]
        return  res.reshape(BS,self.future_steps,-1,self.mul)
       
       
       
          



from torch import  nn
import torch

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from typing import List, Union
from .vva.minigpt import Block
import math
from torch.nn import functional as F
from ..data_structure.utils import beauty_string
from .utils import  get_scope

torch.autograd.set_detect_anomaly(True)

        

class VVA(Base):
    handle_multivariate = False
    handle_future_covariates = False
    handle_categorical_variables = False
    handle_quantile_loss = False
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    
    def __init__(self, 
                 past_steps:int,
                 future_steps:int,
                 past_channels:int,
                 future_channels:int,
                 embs:List[int],
                 d_model:int,
                 max_voc_size:int,
                 token_split: int,
                 num_layers:int,
                 dropout_rate:float,
                 n_heads:int,
                 out_channels:int,
                 persistence_weight:float=0.0,
                 loss_type: str='l1',
                 quantiles:List[int]=[],
                 optim:Union[str,None]=None,
                 optim_config:dict=None,
                 scheduler_config:dict=None,
                 **kwargs)->None:
        """ Custom encoder-decoder 
        
        Args:
            past_steps (int):  number of past datapoints used 
            future_steps (int): number of future lag to predict
            past_channels (int): number of numeric past variables, must be >0
            future_channels (int): number of future numeric variables 
            embs (List): list of the initial dimension of the categorical variables
            cat_emb_dim (int): final dimension of each categorical variable
            hidden_RNN (int): hidden size of the RNN block
            num_layers_RNN (int): number of RNN layers
            kind (str): one among GRU or LSTM
            kernel_size (int): kernel size in the encoder convolutional block
            sum_emb (bool): if true the contribution of each embedding will be summed-up otherwise stacked
            out_channels (int):  number of output channels
            activation (str, optional): activation fuction function pytorch. Default torch.nn.ReLU
            remove_last (bool, optional): if True the model learns the difference respect to the last seen point
            persistence_weight (float):  weight controlling the divergence from persistence model. Default 0
            loss_type (str, optional): this model uses custom losses or l1 or mse. Custom losses can be linear_penalization or exponential_penalization. Default l1,
            quantiles (List[int], optional): we can use quantile loss il len(quantiles) = 0 (usually 0.1,0.5, 0.9) or L1loss in case len(quantiles)==0. Defaults to [].
            dropout_rate (float, optional): dropout rate in Dropout layers
            use_bn (bool, optional): if true BN layers will be added and dropouts will be removed
            use_glu (bool,optional): use GLU for feature selection. Defaults to True.
            glu_percentage (float, optiona): percentage of features to use. Defaults to 1.0.
            n_classes (int): number of classes (0 in regression)
            optim (str, optional): if not None it expects a pytorch optim method. Defaults to None that is mapped to Adam.
            optim_config (dict, optional): configuration for Adam optimizer. Defaults to None.
            scheduler_config (dict, optional): configuration for stepLR scheduler. Defaults to None.

        """

        
        super().__init__(**kwargs)
        self.block_size = past_steps//token_split + future_steps//token_split -1
        self.save_hyperparameters(logger=False)
        self.sentence_length = future_steps//token_split
        
        self.transformer = nn.ModuleDict(dict(
            wte = nn.Embedding(max_voc_size, d_model),
            wpe = nn.Embedding(self.block_size, d_model),
            drop = nn.Dropout(dropout_rate),
            h = nn.ModuleList([Block( d_model,dropout_rate,n_heads,dropout_rate,self.block_size) for _ in range(num_layers)]), ##care can be different dropouts
            ln_f = nn.LayerNorm(d_model),
        ))
        self.lm_head = nn.Linear(d_model, max_voc_size, bias=False)


        for pn, p in self.named_parameters():
            if pn.endswith('c_proj.weight'):
                torch.nn.init.normal_(p, mean=0.0, std=0.02/math.sqrt(2 * num_layers))

        # report number of parameters (note we don't count the decoder parameters in lm_head)
        n_params = sum(p.numel() for p in self.transformer.parameters())
        beauty_string("number of parameters: %.2fM" % (n_params/1e6,),'info',self.verbose)
        
        
        
        
        
        
        
        self.use_quantiles = True
        self.is_classification = True
        self.scheduler_config = scheduler_config
        self.optim_config = optim_config
        self.optim = self.scheduler_config = self.configure_optimizers()

        
    def configure_optimizers(self):
        """
        This long function is unfortunately doing something very simple and is being very defensive:
        We are separating out all parameters of the model into two buckets: those that will experience
        weight decay for regularization and those that won't (biases, and layernorm/embedding weights).
        We are then returning the PyTorch optimizer object.
        """

        # separate out all parameters to those that will and won't experience regularizing weight decay
        decay = set()
        no_decay = set()
        whitelist_weight_modules = (torch.nn.Linear, )
        blacklist_weight_modules = (torch.nn.LayerNorm, torch.nn.Embedding)
        for mn, m in self.named_modules():
            for pn, p in m.named_parameters():
                fpn = '%s.%s' % (mn, pn) if mn else pn # full param name
                # random note: because named_modules and named_parameters are recursive
                # we will see the same tensors p many many times. but doing it this way
                # allows us to know which parent module any tensor p belongs to...
                if pn.endswith('bias'):
                    # all biases will not be decayed
                    no_decay.add(fpn)
                elif pn.endswith('weight') and isinstance(m, whitelist_weight_modules):
                    # weights of whitelist modules will be weight decayed
                    decay.add(fpn)
                elif pn.endswith('weight') and isinstance(m, blacklist_weight_modules):
                    # weights of blacklist modules will NOT be weight decayed
                    no_decay.add(fpn)

        # validate that we considered every parameter
        param_dict = {pn: p for pn, p in self.named_parameters()}
        inter_params = decay & no_decay
        union_params = decay | no_decay
        assert len(inter_params) == 0, beauty_string(f"parameters {inter_params} made it into both decay/no_decay sets!",'section' ,True)
        assert len(param_dict.keys() - union_params) == 0,  beauty_string(f"parameters {param_dict.keys() - union_params} were not separated into either decay/no_decay set!",'section',True)                                            

        # create the pytorch optimizer object
        optim_groups = [
            {"params": [param_dict[pn] for pn in sorted(list(decay))], "weight_decay": self.optim_config.weight_decay},
            {"params": [param_dict[pn] for pn in sorted(list(no_decay))], "weight_decay": 0.0},
        ]
        optimizer = torch.optim.AdamW(optim_groups, lr=self.optim_config.lr, betas=self.optim_config.betas)
        return optimizer

    def compute_loss(self,batch,y_hat):
        """
        custom loss calculation
        
        :meta private:
        """
        return F.cross_entropy(y_hat.view(-1, y_hat.size(-1)), batch['y_emb'].view(-1), ignore_index=-1)


    def forward(self, batch):
        b, t = batch['x_emb'].size()
        assert t <= self.block_size, beauty_string("Cannot forward sequence of length {t}, block size is only {self.block_size}",'section',True)
        pos = torch.arange(0, t, dtype=torch.long, device=self.device).unsqueeze(0) # shape (1, t)

        # forward the GPT model itself
        tok_emb = self.transformer.wte(batch['x_emb']) # token embeddings of shape (b, t, n_embd)
        pos_emb = self.transformer.wpe(pos) # position embeddings of shape (1, t, n_embd)
        x = self.transformer.drop(tok_emb + pos_emb)
        for block in self.transformer.h:
            x = block(x)
        x = self.transformer.ln_f(x)
        logits = self.lm_head(x)

        return logits


    def generate(self, idx, max_new_tokens, temperature=1.0, do_sample=False, top_k=None,num_samples=100):
        """
        Take a conditioning sequence of indices idx (LongTensor of shape (b,t)) and complete
        the sequence max_new_tokens times, feeding the predictions back into the model each time.
        Most likely you'll want to make sure to be in model.eval() mode of operation for this.
        """
        if do_sample:
            idx = idx.repeat(num_samples,1,1)
            for _ in range(max_new_tokens):
                tmp = []
                for i in range(num_samples):
                    idx_cond = idx[i,:,:] if idx.size(2) <= self.block_size else idx[i,:, -self.block_size:]
                    logits = self({'x_emb':idx_cond})
                    logits = logits[:, -1, :] / temperature
                    if top_k is not None:
                        v, _ = torch.topk(logits, top_k)
                        logits[logits < v[:, [-1]]] = -float('Inf')
                    probs = F.softmax(logits, dim=-1)
                    idx_next = torch.multinomial(probs, num_samples=1, replacement=True)
                    tmp.append(idx_next)
                tmp = torch.cat(tmp,dim=1).T.unsqueeze(2)
                idx = torch.cat((idx, tmp), dim=2)
            return idx
        else:
            for _ in range(max_new_tokens):
                
                # if the sequence context is growing too long we must crop it at block_size
                idx_cond = idx if idx.size(1) <= self.block_size else idx[:, -self.block_size:]
                # forward the model to get the logits for the index in the sequence
                logits = self({'x_emb':idx_cond})
                # pluck the logits at the final step and scale by desired temperature
                logits = logits[:, -1, :] / temperature
                # optionally crop the logits to only the top k options
                if top_k is not None:
                    v, _ = torch.topk(logits, top_k)
                    logits[logits < v[:, [-1]]] = -float('Inf')
                # apply softmax to convert logits to (normalized) probabilities
                probs = F.softmax(logits, dim=-1)
                # either sample from the distribution or take the most likely element
                _, idx_next = torch.topk(probs, k=1, dim=-1)
                # append sampled index to the running sequence and continue
                idx = torch.cat((idx, idx_next), dim=1)

            return idx.unsqueeze(0)

    def inference(self, batch:dict)->torch.tensor:
        x = batch['x_emb'].to(self.device)
        
        # isolate the input pattern alone
        inp = x[:, :self.sentence_length]
        
        # let the model sample the rest of the sequence
        cat = self.generate(inp, self.sentence_length, do_sample=True,num_samples=3) # using greedy argmax, not samplingv ##todo here add sampling
        sol_candidate = cat[:,:, self.sentence_length:] 

       
        return sol_candidate.permute(1,2,0)
    



import pandas as pd
import os
import numpy as np
from typing import List, Tuple
import logging
import requests
from bs4 import BeautifulSoup as bs

def build_venice(path:str,url='https://www.comune.venezia.it/it/content/archivio-storico-livello-marea-venezia-1')->None:
    
      
    with requests.Session() as s:
        r = s.get(url)
    soup = bs(r.content)

    print('CARE THE STRUCTURE OF THE SITE CAN BE CHANGED')

    def cast_string(x):
        if np.isfinite(x) is False:
            return x
        if x<10:
            return f'0{int(x)}:00'
        else:
            return f'{int(x)}:00'
        
    def cast_month(x):
        try:
            return x.replace('gen','01').replace('feb','02').replace('mar','03').replace('apr','04').replace('mag','05').replace('giu','06').replace('lug','07').replace('ago','08').replace('set','09').replace('ott','10').replace('nov','11').replace('dic','12')
        except:
            return x
        
    def remove_float(table,column):
        if table[column].dtype in [int,float]:
            table[column] = table[column].apply(lambda x:cast_string(x))
        else:
            pass
        
    def remove_str(table,column):
        table[column] = table[column].apply(lambda x:cast_month(x))
            
    def normalize(table):
        columns = table.columns
        table = table[~table.isna().all(axis=1)]
        if 'Data_ora(solare)' in columns:
            table['time'] = table['Data_ora(solare)'] 
        
        elif 'GIORNO' in columns and 'ORA solare' in columns:
            remove_float(table,'ORA solare')
            table['time'] = table['GIORNO'] +' '+ table['ORA solare'] 
        
        elif 'data' in columns and 'ora solare' in columns:
            remove_float(table,'ora solare')
            table['time'] =table['data'] +' '+ table['ora solare'] 
        
        elif 'Data' in columns and 'Ora solare' in columns:
            remove_str(table,'Data')
            remove_float(table,'Ora solare')
            table['time'] = table['Data'] +' '+ table['Ora solare'] 
        elif 'GIORNO' in columns and 'ORA' in columns:
            remove_str(table,'GIORNO')
            remove_float(table,'ORA')
            table['time'] = table['GIORNO'] +' '+ table['ORA'] 
        
        else:
            import pdb
            pdb.set_trace()
       
        for c in columns:
            if 'Salute' in c:
                table['y'] = table[c].values
                if 'cm' in c:
                    table['y']/=100
        res = table[['time','y']].dropna()
        try:
            res['time'] = pd.to_datetime(res['time'],format='mixed')
        except:
            import pdb
            pdb.set_trace()
        return res
    tot= []
    for row in soup.find_all("table")[1].find('tbody').find_all('tr'):
        for i,column in enumerate(row.find_all('td')):
            tmp_links = column.find_all('a')
            if len(tmp_links)>0:
                for x in tmp_links:
                    if 'orari' in x['href']:
                        tmp =  pd.read_csv('https://www.comune.venezia.it/'+x['href'],sep=';', parse_dates=True)
                        tot.append(normalize(tmp))
              
    
    res = pd.concat(tot)
    res.sort_values(by='time',inplace=True)
    res.to_csv(f'{path}/venice.csv',index=False)



def read_public_dataset(path:str,dataset:str)->Tuple[pd.DataFrame,List[str]]:
    """
    Returns the public dataset chosen. Pleas download the dataset from here https://drive.google.com/drive/folders/1ZOYpTUa82_jCcxIdTmyr0LXQfvaM9vIy or ask to agobbi@fbk.eu. 


    Args:
        path (str): path to data
        dataset (str): dataset (one of 'electricity','etth1','etth2','ettm1','ettm2','exchange_rate','illness','traffic','weather')

    Returns:
        Tuple[pd.DataFrame,List[str]]: The target variable is *y* and the time index is *time* and the list of the covariates
    """
    
    

    if os.path.isdir(path):
        pass
    else:
        logging.info('I will try to create the folder')
        os.mkdir(path)
        
    files = os.listdir(path)
    if 'all_six_datasets' in files:
        pass
    else:
        logging.error('Please dowload the zip file form here and unzip it https://drive.google.com/drive/folders/1ZOYpTUa82_jCcxIdTmyr0LXQfvaM9vIy')
        return None,None
    
    
    if dataset not in ['electricity','etth1','etth2','ettm1','ettm2','exchange_rate','illness','traffic','weather','venice']:
        logging.error(f"Dataset {dataset} not available, use one among ['electricity','etth1','etth2','ettm1','ettm2','exchange_rate','illness','traffic','weather','venice']")
        return None,None

    if dataset=='electricity':
        dataset = pd.read_csv(os.path.join(path,'all_six_datasets/electricity/electricity.csv'),sep=',',na_values=-9999)
    elif dataset=='etth1':
        dataset = pd.read_csv(os.path.join(path,'all_six_datasets/ETT-small/ETTh1.csv'),sep=',',na_values=-9999)  
    elif dataset=='etth2':
        dataset = pd.read_csv(os.path.join(path,'all_six_datasets/ETT-small/ETTh2.csv'),sep=',',na_values=-9999)
    elif dataset=='ettm1':
        dataset = pd.read_csv(os.path.join(path,'all_six_datasets/ETT-small/ETTm1.csv'),sep=',',na_values=-9999)
    elif dataset=='ettm2':
        dataset = pd.read_csv(os.path.join(path,'all_six_datasets/ETT-small/ETTm2.csv'),sep=',',na_values=-9999)
    elif dataset=='exchange_rate':
        dataset = pd.read_csv(os.path.join(path,'all_six_datasets/exchange_rate/exchange_rate.csv'),sep=',',na_values=-9999)
    elif dataset=='illness':
        dataset = pd.read_csv(os.path.join(path,'all_six_datasets/illness/national_illness.csv'),sep=',',na_values=-9999)
    elif dataset=='traffic':
        dataset = pd.read_csv(os.path.join(path,'all_six_datasets/traffic/traffic.csv'),sep=',',na_values=-9999) 
    elif dataset=='weather':
        dataset = pd.read_csv(os.path.join(path,'all_six_datasets/weather/weather.csv'),sep=',',na_values=-9999) 
    elif dataset=='venice':
        if os.path.isfile(os.path.join(path,'venice.csv')):
            dataset = pd.read_csv(os.path.join(path,'venice.csv')) 
        else:
            logging.info('I WILL TRY TO DOWNLOAD IT, if there are errors please have a look to `build_venice` function')
            build_venice(path,url='https://www.comune.venezia.it/it/content/archivio-storico-livello-marea-venezia-1')
            dataset = pd.read_csv(os.path.join(path,'venice.csv')) 
    else:
        logging.error(f'Dataset {dataset} not found')
        return None, None
    dataset.rename(columns={'date':'time','OT':'y'},inplace=True)
    dataset.time = pd.to_datetime(dataset.time)
    logging.info(f'Dataset loaded with shape {dataset.shape}')
    
    return dataset, list(set(dataset.columns).difference(set(['time','y'])))


## Copyright https://github.com/romilbert/samformer/tree/main?tab=MIT-1-ov-file#readme
## Modified for notation alignmenet and batch structure
## extended to what inside samformer folder

import torch
import torch.nn as nn
import numpy as np
from .samformer.utils import scaled_dot_product_attention, RevIN



try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from .utils import QuantileLossMO,Permute, get_activation

from typing import List, Union
from ..data_structure.utils import beauty_string
from .utils import  get_scope
from .utils import Embedding_cat_variables




class Samformer(Base):
    handle_multivariate = True
    handle_future_covariates = False # or at least it seems...
    handle_categorical_variables = False #solo nel encoder
    handle_quantile_loss = False # NOT EFFICIENTLY ADDED, TODO fix this
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    def __init__(self, 
               
                 # specific params
                 hidden_size:int,
                 use_revin: bool,
                 activation: str='',
                 
                 **kwargs)->None:
        """Initialize the model with specified parameters. Samformer: Unlocking the Potential of Transformers in Time Series Forecasting with Sharpness-Aware Minimization and Channel-Wise Attention.
        https://arxiv.org/pdf/2402.10198
        
        Args:
            hidden_size (int): The size of the hidden layer.
            use_revin (bool): Flag indicating whether to use RevIN.
            activation (str, optional): The activation function to use. Defaults to ''.
            **kwargs: Additional keyword arguments passed to the parent class.
        
        Raises:
            ValueError: If the activation function is not recognized.
        

        """
        
        super().__init__(**kwargs)
        if activation == 'torch.nn.SELU':
            beauty_string('SELU do not require BN','info',self.verbose)
            use_bn = False
        if isinstance(activation,str):
            activation = get_activation(activation)
        self.save_hyperparameters(logger=False)


        self.emb_past = Embedding_cat_variables(self.past_steps,self.emb_dim,self.embs_past, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        self.emb_fut = Embedding_cat_variables(self.future_steps,self.emb_dim,self.embs_fut, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        emb_past_out_channel = self.emb_past.output_channels
        emb_fut_out_channel = self.emb_fut.output_channels
        
        
        self.revin = RevIN(num_features=self.past_channels+emb_past_out_channel)
        self.compute_keys = nn.Linear(self.past_steps, hidden_size)
        self.compute_queries = nn.Linear(self.past_steps, hidden_size)
        self.compute_values = nn.Linear(self.past_steps, self.past_steps)
        self.linear_forecaster = nn.Linear(self.past_steps, self.future_steps)
        self.use_revin = use_revin

        dim = emb_past_out_channel+self.past_channels+emb_fut_out_channel+self.future_channels
        self.final_layer = nn.Sequential(activation(),
                                         nn.Linear(dim, dim*2),
                                         activation(),
                                         nn.Linear(dim*2,self.out_channels*self.mul  ))

    def can_be_compiled(self):
        return True  
  
    def forward(self, batch:dict)-> float:

        x = batch['x_num_past'].to(self.device)
        BS = x.shape[0]
        if 'x_cat_future' in batch.keys():
            emb_fut = self.emb_fut(BS,batch['x_cat_future'].to(self.device))
        else:
            emb_fut = self.emb_fut(BS,None)
        if 'x_cat_past' in batch.keys():
            emb_past = self.emb_past(BS,batch['x_cat_past'].to(self.device))
        else:
            emb_past = self.emb_past(BS,None)
            
        tmp_future = [emb_fut]
        if 'x_num_future' in batch.keys():
            x_future = batch['x_num_future'].to(self.device)
            tmp_future.append(x_future)
            
        tot = [x,emb_past]
        x = torch.cat(tot,axis=2)

        
 

        if self.use_revin:
            x_norm = self.revin(x, mode='norm').transpose(1, 2) # (n, D, L)
        else:
            x_norm = x.transpose(1, 2)
        # Channel-Wise Attention

        queries = self.compute_queries(x_norm) # (n, D, hid_dim)
        keys = self.compute_keys(x_norm) # (n, D, hid_dim)
        values = self.compute_values(x_norm) # (n, D, L)
        if hasattr(nn.functional, 'scaled_dot_product_attention'):
            att_score = nn.functional.scaled_dot_product_attention(queries, keys, values) # (n, D, L)
        else:
            att_score = scaled_dot_product_attention(queries, keys, values) # (n, D, L)
        out = x_norm + att_score # (n, D, L)
        # Linear Forecasting
        out = self.linear_forecaster(out) # (n, D, H)
        # RevIN Denormalization
        if self.use_revin:
            out = self.revin(out.transpose(1, 2), mode='denorm').transpose(1, 2) # (n, D, H)


        tmp_future.append(out.permute(0,2,1))
        tmp_future = torch.cat(tmp_future,2)
        output = self.final_layer(tmp_future)

        return output.reshape(BS,self.future_steps,self.out_channels,self.mul)





from torch import nn

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from .utils import L1Loss
from ..data_structure.utils import beauty_string
from .utils import  get_scope

class Persistent(Base):
    handle_multivariate = True
    handle_future_covariates = False
    handle_categorical_variables = False
    handle_quantile_loss = False
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    def __init__(self, 
                 **kwargs)->None:
        """
        Simple persistent model aligned with all the other
        """
    
        super().__init__(**kwargs)
        self.save_hyperparameters(logger=False)
        self.fake = nn.Linear(1,1)
        self.use_quantiles = False
    def can_be_compiled(self):
        return False
    
    def forward(self, batch):
     
        x =  batch['x_num_past'].to(self.device)
        idx_target = batch['idx_target'][0]
        x_start = x[:,-1,idx_target].unsqueeze(1)
        #this is B,1,C
        #[B,L,C,1] remember the outoput size
        res = x_start.repeat(1,self.future_steps,1).unsqueeze(3)
        
        return res
    


import torch.nn as nn
from einops import rearrange
from .attn import  AttentionLayer, TwoStageAttentionLayer

class DecoderLayer(nn.Module):
    '''
    The decoder layer of Crossformer, each layer will make a prediction at its scale
    '''
    def __init__(self, seg_len, d_model, n_heads, d_ff=None, dropout=0.1, out_seg_num = 10, factor = 10):
        super(DecoderLayer, self).__init__()
        self.self_attention = TwoStageAttentionLayer(out_seg_num, factor, d_model, n_heads,d_ff, dropout)
        self.cross_attention = AttentionLayer(d_model, n_heads, dropout = dropout)
        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)
        self.MLP1 = nn.Sequential(nn.Linear(d_model, d_model),
                                nn.GELU(),
                                nn.Linear(d_model, d_model))
        self.linear_pred = nn.Linear(d_model, seg_len)

    def forward(self, x, cross):
        '''
        x: the output of last decoder layer
        cross: the output of the corresponding encoder layer
        '''

        batch = x.shape[0]
        x = self.self_attention(x)
        x = rearrange(x, 'b ts_d out_seg_num d_model -> (b ts_d) out_seg_num d_model')
        
        cross = rearrange(cross, 'b ts_d in_seg_num d_model -> (b ts_d) in_seg_num d_model')
        tmp = self.cross_attention(
            x, cross, cross,
        )
        x = x + self.dropout(tmp)
        y = x = self.norm1(x)
        y = self.MLP1(y)
        dec_output = self.norm2(x+y)
        
        dec_output = rearrange(dec_output, '(b ts_d) seg_dec_num d_model -> b ts_d seg_dec_num d_model', b = batch)
        layer_predict = self.linear_pred(dec_output)
        layer_predict = rearrange(layer_predict, 'b out_d seg_num seg_len -> b (out_d seg_num) seg_len')

        return dec_output, layer_predict

class Decoder(nn.Module):
    '''
    The decoder of Crossformer, making the final prediction by adding up predictions at each scale
    '''
    def __init__(self, seg_len, d_layers, d_model, n_heads, d_ff, dropout,\
                router=False, out_seg_num = 10, factor=10):
        super(Decoder, self).__init__()

        self.router = router
        self.decode_layers = nn.ModuleList()
        for i in range(d_layers):
            self.decode_layers.append(DecoderLayer(seg_len, d_model, n_heads, d_ff, dropout, \
                                        out_seg_num, factor))

    def forward(self, x, cross):
        final_predict = None
        i = 0

        ts_d = x.shape[1]
        for layer in self.decode_layers:
            cross_enc = cross[i]
            x, layer_predict = layer(x,  cross_enc)
            if final_predict is None:
                final_predict = layer_predict
            else:
                final_predict = final_predict + layer_predict
            i += 1
        
        final_predict = rearrange(final_predict, 'b (out_d seg_num) seg_len -> b (seg_num seg_len) out_d', out_d = ts_d)

        return final_predict




import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import logging

'''
class VectorQuantizer(nn.Module):
    """
    Inspired from Sonnet implementation of VQ-VAE https://arxiv.org/abs/1711.00937,
    in https://github.com/deepmind/sonnet/blob/master/sonnet/python/modules/nets/vqvae.py and
    pytorch implementation of it from zalandoresearch in https://github.com/zalandoresearch/pytorch-vq-vae/blob/master/vq-vae.ipynb.

    Implements the algorithm presented in
    'Neural Discrete Representation Learning' by van den Oord et al.
    https://arxiv.org/abs/1711.00937

    Input any tensor to be quantized. Last dimension will be used as space in
    which to quantize. All other dimensions will be flattened and will be seen
    as different examples to quantize.
    The output tensor will have the same shape as the input.
    For example a tensor with shape [16, 32, 32, 64] will be reshaped into
    [16384, 64] and all 16384 vectors (each of 64 dimensions)  will be quantized
    independently.
    Args:
        embedding_dim: integer representing the dimensionality of the tensors in the
            quantized space. Inputs to the modules must be in this format as well.
        num_embeddings: integer, the number of vectors in the quantized space.
            commitment_cost: scalar which controls the weighting of the loss terms
            (see equation 4 in the paper - this variable is Beta).
    """
    
    def __init__(self, num_embeddings, embedding_dim, commitment_cost, device):
        super(VectorQuantizer, self).__init__()

        self._embedding_dim = embedding_dim
        self._num_embeddings = num_embeddings
        
        self._embedding = nn.Embedding(self._num_embeddings, self._embedding_dim)
        self._embedding.weight.data.uniform_(-1/self._num_embeddings, 1/self._num_embeddings)

        self._commitment_cost = commitment_cost
        self._device = device

    def forward(self, inputs, compute_distances_if_possible=True, record_codebook_stats=False):
        """
        Connects the module to some inputs.

        Args:
            inputs: Tensor, final dimension must be equal to embedding_dim. All other
                leading dimensions will be flattened and treated as a large batch.

        Returns:
            loss: Tensor containing the loss to optimize.
            quantize: Tensor containing the quantized version of the input.
            perplexity: Tensor containing the perplexity of the encodings.
            encodings: Tensor containing the discrete encodings, ie which element
                of the quantized space each input element was mapped to.
            distances
        """

        # Convert inputs from BCHW -> BHWC
        inputs = inputs.permute(1, 2, 0).contiguous()
        input_shape = inputs.shape
        _, time, batch_size = input_shape

        # Flatten input
        flat_input = inputs.view(-1, self._embedding_dim)

        # Compute distances between encoded audio frames and embedding vectors
        distances = (torch.sum(flat_input**2, dim=1, keepdim=True)
            + torch.sum(self._embedding.weight**2, dim=1)
            - 2 * torch.matmul(flat_input, self._embedding.weight.t()))

        """
        encoding_indices: Tensor containing the discrete encoding indices, ie
        which element of the quantized space each input element was mapped to.
        """
        encoding_indices = torch.argmin(distances, dim=1).unsqueeze(1)
        encodings = torch.zeros(encoding_indices.shape[0], self._num_embeddings, dtype=torch.float).to(self._device)
        encodings.scatter_(1, encoding_indices, 1)

        # Compute distances between encoding vectors
        if not self.training and compute_distances_if_possible:
            _encoding_distances = [torch.dist(items[0], items[1], 2).to(self._device) for items in combinations(flat_input, r=2)]
            encoding_distances = torch.tensor(_encoding_distances).to(self._device).view(batch_size, -1)
        else:
            encoding_distances = None

        # Compute distances between embedding vectors
        if not self.training and compute_distances_if_possible:
            _embedding_distances = [torch.dist(items[0], items[1], 2).to(self._device) for items in combinations(self._embedding.weight, r=2)]
            embedding_distances = torch.tensor(_embedding_distances).to(self._device)
        else:
            embedding_distances = None

        # Sample nearest embedding
        if not self.training and compute_distances_if_possible:
            _frames_vs_embedding_distances = [torch.dist(items[0], items[1], 2).to(self._device) for items in product(flat_input, self._embedding.weight.detach())]
            frames_vs_embedding_distances = torch.tensor(_frames_vs_embedding_distances).to(self._device).view(batch_size, time, -1)
        else:
            frames_vs_embedding_distances = None

        # Quantize and unflatten
        quantized = torch.matmul(encodings, self._embedding.weight).view(input_shape)
        # TODO: Check if the more readable self._embedding.weight.index_select(dim=1, index=encoding_indices) works better

        concatenated_quantized = self._embedding.weight[torch.argmin(distances, dim=1).detach().cpu()] if not self.training or record_codebook_stats else None

        # Losses
        e_latent_loss = torch.mean((quantized.detach() - inputs)**2)
        q_latent_loss = torch.mean((quantized - inputs.detach())**2)
        commitment_loss = self._commitment_cost * e_latent_loss
        vq_loss = q_latent_loss + commitment_loss

        quantized = inputs + (quantized - inputs).detach() # Trick to prevent backpropagation of quantized
        avg_probs = torch.mean(encodings, dim=0)

        """
        The perplexity a useful value to track during training.
        It indicates how many codes are 'active' on average.
        """
        perplexity = torch.exp(-torch.sum(avg_probs * torch.log(avg_probs + 1e-10))) # Exponential entropy

        # Convert quantized from BHWC -> BCHW
        return vq_loss, quantized.permute(2, 0, 1).contiguous(), \
            perplexity, encodings.view(batch_size, time, -1), \
            distances.view(batch_size, time, -1), encoding_indices, \
            {'e_latent_loss': e_latent_loss.item(), 'q_latent_loss': q_latent_loss.item(),
            'commitment_loss': commitment_loss.item(), 'vq_loss': vq_loss.item()}, \
            encoding_distances, embedding_distances, frames_vs_embedding_distances, concatenated_quantized

    @property
    def embedding(self):
        return self._embedding
'''

class VectorQuantizer(nn.Module):
    def __init__(self, num_embeddings, embedding_dim, commitment_cost):
        super(VectorQuantizer, self).__init__()
        
        self._embedding_dim = embedding_dim
        self._num_embeddings = num_embeddings
        
        self._embedding = nn.Embedding(self._num_embeddings, self._embedding_dim)
        self._embedding.weight.data.uniform_(-1/self._num_embeddings, 1/self._num_embeddings)
        self._commitment_cost = commitment_cost

    def forward(self, inputs):
        # convert inputs from BCHW -> BHWC
        inputs = inputs.permute(0, 2, 1).contiguous()
        input_shape = inputs.shape
        
        # Flatten input
        flat_input = inputs.view(-1, self._embedding_dim)
        
        # Calculate distances
        distances = (torch.sum(flat_input**2, dim=1, keepdim=True) 
                    + torch.sum(self._embedding.weight**2, dim=1)
                    - 2 * torch.matmul(flat_input, self._embedding.weight.t()))
            
        # Encoding
        encoding_indices = torch.argmin(distances, dim=1).unsqueeze(1)
        encodings = torch.zeros(encoding_indices.shape[0], self._num_embeddings, device=inputs.device)
        encodings.scatter_(1, encoding_indices, 1)
        
        # Quantize and unflatten
        quantized = torch.matmul(encodings, self._embedding.weight).view(input_shape)
        
        # Loss
        e_latent_loss = F.mse_loss(quantized.detach(), inputs)
        q_latent_loss = F.mse_loss(quantized, inputs.detach())
        loss = q_latent_loss + self._commitment_cost * e_latent_loss
        
        quantized = inputs + (quantized - inputs).detach()
        avg_probs = torch.mean(encodings, dim=0)
        perplexity = torch.exp(-torch.sum(avg_probs * torch.log(avg_probs + 1e-10)))
        
        # convert quantized from BHWC -> BCHW
        #return loss, quantized.permute(0, 3, 1, 2).contiguous(), perplexity, encodings
        return loss, quantized.permute(0, 2, 1).contiguous(), perplexity, encodings.view(input_shape[0],-1,encodings.shape[1])

class VectorQuantizerEMA(nn.Module):
    
    def __init__(self, num_embeddings, embedding_dim, commitment_cost, decay, epsilon=1e-5):
        super(VectorQuantizerEMA, self).__init__()
        
        self._embedding_dim = embedding_dim
        self._num_embeddings = num_embeddings
        
        self._embedding = nn.Embedding(self._num_embeddings, self._embedding_dim)
        self._embedding.weight.data.normal_()
        self._commitment_cost = commitment_cost
        
        self.register_buffer('_ema_cluster_size', torch.zeros(num_embeddings))
        self._ema_w = nn.Parameter(torch.Tensor(num_embeddings, self._embedding_dim))
        self._ema_w.data.normal_()
        
        self._decay = decay
        self._epsilon = epsilon

    def forward(self, inputs):
        # convert inputs from BCHW -> BHWC
        inputs = inputs.permute(0, 2, 1).contiguous()
        input_shape = inputs.shape
        
        # Flatten input
        flat_input = inputs.view(-1, self._embedding_dim)
        
        # Calculate distances
        distances = (torch.sum(flat_input**2, dim=1, keepdim=True) 
                    + torch.sum(self._embedding.weight**2, dim=1)
                    - 2 * torch.matmul(flat_input, self._embedding.weight.t()))
            
        # Encoding
        encoding_indices = torch.argmin(distances, dim=1).unsqueeze(1)
        encodings = torch.zeros(encoding_indices.shape[0], self._num_embeddings, device=inputs.device)
        encodings.scatter_(1, encoding_indices, 1)
        
        # Quantize and unflatten
        quantized = torch.matmul(encodings, self._embedding.weight).view(input_shape)
        
        # Use EMA to update the embedding vectors
        if self.training:
            self._ema_cluster_size = self._ema_cluster_size * self._decay + \
                                     (1 - self._decay) * torch.sum(encodings, 0)
            
            # Laplace smoothing of the cluster size
            n = torch.sum(self._ema_cluster_size.data)
            self._ema_cluster_size = (
                (self._ema_cluster_size + self._epsilon)
                / (n + self._num_embeddings * self._epsilon) * n)
            
            dw = torch.matmul(encodings.t(), flat_input)
            self._ema_w = nn.Parameter(self._ema_w * self._decay + (1 - self._decay) * dw)
            
            self._embedding.weight = nn.Parameter(self._ema_w / self._ema_cluster_size.unsqueeze(1))
        
        # Loss
        e_latent_loss = F.mse_loss(quantized.detach(), inputs)
        loss = self._commitment_cost * e_latent_loss
        
        # Straight Through Estimator
        quantized = inputs + (quantized - inputs).detach()
        avg_probs = torch.mean(encodings, dim=0)
        perplexity = torch.exp(-torch.sum(avg_probs * torch.log(avg_probs + 1e-10)))
        
        # convert quantized from BHWC -> BCHW
        return loss, quantized.permute(0, 2, 1).contiguous(), perplexity, encodings.view(input_shape[0],-1,encodings.shape[1])
    
    
class Residual(nn.Module):

    def __init__(self, in_channels, hidden_channels, num_residual_hiddens):
        super(Residual, self).__init__()
        
        relu_1 = nn.ReLU(True)
        conv_1 = nn.Conv1d(
            in_channels=in_channels,
            out_channels=num_residual_hiddens,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False
        )


        relu_2 = nn.ReLU(True)
        conv_2 = nn.Conv1d(
            in_channels=num_residual_hiddens,
            out_channels=hidden_channels,
            kernel_size=1,
            stride=1,
            bias=False
        )


        # All parameters same as specified in the paper
        self._block = nn.Sequential(
            relu_1,
            conv_1,
            relu_2,
            conv_2
        )
    
    def forward(self, x):
        return x + self._block(x)
    
class ResidualStack(nn.Module):

    def __init__(self, in_channels, hidden_channels, num_residual_layers, num_residual_hiddens):
        super(ResidualStack, self).__init__()
        
        self._num_residual_layers = num_residual_layers
        self._layers = nn.ModuleList([Residual(in_channels, hidden_channels, num_residual_hiddens)] * self._num_residual_layers)
        
    def forward(self, x):
        for i in range(self._num_residual_layers):
            x = self._layers[i](x)
        return F.relu(x)
    
class Encoder(nn.Module):
    def __init__(self, in_channels, hidden_channels,num_residual_layers=3):
        super(Encoder, self).__init__()

        self._conv_1 = nn.Conv1d(in_channels=in_channels,
                                 out_channels=hidden_channels,
                                 kernel_size=3, padding=1)
        
        self._conv_2 = nn.Conv1d(in_channels=hidden_channels,
                                 out_channels=hidden_channels,
                                 kernel_size=3,padding=1)

        self._conv_3 = nn.Conv1d(in_channels=hidden_channels,
                                 out_channels=hidden_channels,
                                 kernel_size=4,
                                 stride=2, padding=1)
        
        self._conv_4 = nn.Conv1d(in_channels=hidden_channels,
                                 out_channels=hidden_channels,
                                 kernel_size=3,
                                 padding=1)
        
        self._conv_5 = nn.Conv1d(in_channels=hidden_channels,
                                 out_channels=hidden_channels,
                                 kernel_size=3, padding=1)

        self._residual_stack = ResidualStack(
                in_channels=hidden_channels,
                hidden_channels=hidden_channels,
                num_residual_layers=num_residual_layers,
                num_residual_hiddens=hidden_channels//2
        )
    def forward(self, inputs):
        x_conv_1 = F.relu(self._conv_1(inputs))
        x = F.relu(self._conv_2(x_conv_1)) + x_conv_1
        x_conv_3 = F.relu(self._conv_3(x))
        x_conv_4 = F.relu(self._conv_4(x_conv_3)) + x_conv_3
        x_conv_5 = F.relu(self._conv_5(x_conv_4)) + x_conv_4
        x = self._residual_stack(x_conv_5) + x_conv_5
        

        return x
    
    

class Jitter(nn.Module):
    """
    Jitter implementation from [Chorowski et al., 2019].
    During training, each latent vector can replace either one or both of
    its neighbors. As in dropout, this prevents the model from
    relying on consistency across groups of tokens. Additionally,
    this regularization also promotes latent representation stability
    over time: a latent vector extracted at time step t must strive
    to also be useful at time steps t − 1 or t + 1.
    """

    def __init__(self, probability=0.12):
        super(Jitter, self).__init__()

        self._probability = probability

    def forward(self, quantized):
        original_quantized = quantized.detach().clone()
        length = original_quantized.size(2)
        for i in range(length):
            """
            Each latent vector is replace with either of its neighbors with a certain probability
            (0.12 from the paper).
            """
            replace = [True, False][np.random.choice([1, 0], p=[self._probability, 1 - self._probability])]
            if replace:
                if i == 0:
                    neighbor_index = i + 1
                elif i == length - 1:
                    neighbor_index = i - 1
                else:
                    """
                    "We independently sample whether it is to
                    be replaced with the token right after
                    or before it."
                    """
                    neighbor_index = i + np.random.choice([-1, 1], p=[0.5, 0.5])
                quantized[:, :, i] = original_quantized[:, :, neighbor_index]

        return quantized
    
class Decoder(nn.Module):
    def __init__(self, in_channels, hidden_channels,out_channels,num_residual_layers=3):
        super(Decoder, self).__init__()
        
        
        self._jitter = Jitter(0.125)
        
        self._conv_1 = nn.Conv1d(in_channels=in_channels,
                                 out_channels=hidden_channels,
                                 kernel_size=3, 
                                 padding=1)
        
        self._upsample = nn.Upsample(scale_factor=2)
        
        self._residual_stack = ResidualStack(
            in_channels=hidden_channels,
            hidden_channels=hidden_channels,
            num_residual_layers=num_residual_layers,
            num_residual_hiddens=hidden_channels//2
        )
        
        self._conv_trans_1 = nn.ConvTranspose1d(in_channels=hidden_channels, 
                                                out_channels=hidden_channels,
                                                kernel_size=3,  padding=1)
        
        self._conv_trans_2 = nn.ConvTranspose1d(
                                                in_channels=hidden_channels, 
                                                out_channels=hidden_channels,
                                                kernel_size=4,padding=2)
        
        self._conv_trans_3 = nn.ConvTranspose1d(in_channels=hidden_channels, 
                                                out_channels=out_channels,
                                                kernel_size=4,  padding=1)

    def forward(self, x,is_training=True):
        #if is_training:
        #     x = self._jitter(x)
        x = self._conv_1(x)
        x = self._upsample(x)
        x = self._residual_stack(x)
        x = F.relu(self._conv_trans_1(x))
        x = F.relu(self._conv_trans_2(x))
        x = self._conv_trans_3(x)
        return x
    
class VQVAE(nn.Module):
    def __init__(self,in_channels, hidden_channels, out_channels,num_embeddings, embedding_dim, commitment_cost, decay):
        super(VQVAE, self).__init__()
        
        self._encoder = Encoder(in_channels, hidden_channels)
        self._pre_vq_conv = nn.Conv1d(in_channels=hidden_channels, 
                                      out_channels=embedding_dim,
                                      kernel_size=1, 
                                      stride=1)
        if decay > 0.0:
            self._vq_vae = VectorQuantizerEMA(num_embeddings, embedding_dim, 
                                              commitment_cost, decay)
        else:
            logging.info('CARE NOT TESTED')
            self._vq_vae = VectorQuantizer(num_embeddings, embedding_dim,
                                           commitment_cost)
        self._decoder = Decoder(in_channels=embedding_dim,hidden_channels=hidden_channels,out_channels=out_channels)


    def forward(self, x,is_training=True):
        z = self._encoder(x)

        z = self._pre_vq_conv(z)
        loss, quantized, perplexity, encodings = self._vq_vae(z)
        x_recon = self._decoder(quantized,is_training)

        return loss, x_recon, perplexity,quantized,encodings



## Copyright 2025    DUET (https://github.com/decisionintelligence/DUET)
## Code modified for align the notation and the batch generation
## extended to all present in duet and autoformer folder

import torch
import torch.nn as nn
import numpy as np

from .duet.layers import Linear_extractor_cluster
from .duet.masked import Mahalanobis_mask, Encoder, EncoderLayer, FullAttention, AttentionLayer
from einops import rearrange

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from .utils import QuantileLossMO,Permute, get_activation

from typing import List, Union
from ..data_structure.utils import beauty_string
from .utils import  get_scope
from .utils import Embedding_cat_variables




class Duet(Base):
    handle_multivariate = True
    handle_future_covariates = True 
    handle_categorical_variables = True
    handle_quantile_loss = True 
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    def __init__(self, 
                 
                 factor:int,
                 d_model: int,
                 n_head: int,
                 n_layer: int,
                 CI: bool,
                 d_ff: int,
                 noisy_gating:bool,
                 num_experts: int,
                 kernel_size:int,
                 hidden_size:int,
                 k: int,
                 dropout_rate: float=0.1,
                 activation: str='',
                 **kwargs)->None:
        """Initializes the model with the specified parameters. https://github.com/decisionintelligence/DUET
        
        Args:
            factor (int): The factor for attention scaling. NOT USED but in the original implementation
            d_model (int): The dimensionality of the model.
            n_head (int): The number of attention heads.
            n_layer (int): The number of layers in the encoder.
            CI (bool): Perform channel independent operations.
            d_ff (int): The dimensionality of the feedforward layer.
            noisy_gating (bool): Flag to indicate if noisy gating is used.
            num_experts (int): The number of experts in the mixture of experts.
            kernel_size (int): The size of the convolutional kernel.
            hidden_size (int): The size of the hidden layer.
            k (int): The number of clusters for the linear extractor.
            dropout_rate (float, optional): The dropout rate. Defaults to 0.1.
            activation (str, optional): The activation function to use. Defaults to ''.
            **kwargs: Additional keyword arguments.
        
        Raises:
            ValueError: If the activation function is not recognized.

        """
        
        super().__init__(**kwargs)
        if activation == 'torch.nn.SELU':
            beauty_string('SELU do not require BN','info',self.verbose)
            use_bn = False
        if isinstance(activation,str):
            activation = get_activation(activation)
        self.save_hyperparameters(logger=False)


        self.emb_past = Embedding_cat_variables(self.past_steps,self.emb_dim,self.embs_past, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        self.emb_fut = Embedding_cat_variables(self.future_steps,self.emb_dim,self.embs_fut, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        emb_past_out_channel = self.emb_past.output_channels
        emb_fut_out_channel = self.emb_fut.output_channels


       


        self.cluster = Linear_extractor_cluster(noisy_gating,
                                                num_experts,
                                                self.past_steps,
                                                k,
                                                d_model,
                                                self.past_channels+emb_past_out_channel,
                                                CI,kernel_size,
                                                hidden_size)
        self.CI = CI
        self.n_vars = self.out_channels
        self.mask_generator = Mahalanobis_mask(self.future_steps)
        self.Channel_transformer = Encoder(
            [
                EncoderLayer(
                    AttentionLayer(
                        FullAttention(
                            True,
                            factor,
                            attention_dropout=dropout_rate,
                            output_attention=0,
                        ),
                        d_model,
                        n_head,
                    ),
                    d_model,
                    d_ff,
                    dropout=dropout_rate,
                    activation=activation,
                )
                for _ in range(n_layer)
            ],
            norm_layer=torch.nn.LayerNorm(d_model)
        )

        self.linear_head = nn.Sequential(nn.Linear(d_model, self.future_steps), nn.Dropout(dropout_rate))


        dim = self.past_channels+emb_past_out_channel+emb_fut_out_channel+self.future_channels
        self.final_layer = nn.Sequential(activation(),
                                         nn.Linear(dim, dim*2),
                                         activation(),
                                         nn.Linear(dim*2,self.out_channels*self.mul  ))

    def can_be_compiled(self):
        return False  
    def forward(self, batch:dict)-> float:
        # x: [Batch, Input length, Channel]
        x_enc = batch['x_num_past'].to(self.device)
        idx_target = batch['idx_target'][0]
        BS = x_enc.shape[0]
        
        if 'x_cat_past' in batch.keys():
            emb_past = self.emb_past(BS,batch['x_cat_past'].to(self.device))
        else:
            emb_past = self.emb_past(BS,None)

        
        if 'x_cat_future' in batch.keys():
            emb_fut = self.emb_fut(BS,batch['x_cat_future'].to(self.device))
        else:
            emb_fut = self.emb_fut(BS,None)  
            
        tmp_future = [emb_fut]
        if 'x_num_future' in batch.keys():
            x_future = batch['x_num_future'].to(self.device)
            tmp_future.append(x_future)
      
        x_enc = torch.concat([x_enc,emb_past],axis=-1)
        
        if self.CI:
            channel_independent_input = rearrange(x_enc, 'b l n -> (b n) l 1')

            reshaped_output, _ = self.cluster(channel_independent_input)

            temporal_feature = rearrange(reshaped_output, '(b n) l 1 -> b l n', b=x_enc.shape[0])

        else:
            temporal_feature, _ = self.cluster(x_enc)

        # B x d_model x n_vars -> B x n_vars x d_model
        temporal_feature = rearrange(temporal_feature, 'b d n -> b n d')
        if self.n_vars > 1:
            changed_input = rearrange(x_enc, 'b l n -> b n l')
            channel_mask = self.mask_generator(changed_input)

            channel_group_feature, _ = self.Channel_transformer(x=temporal_feature, attn_mask=channel_mask)

            output = self.linear_head(channel_group_feature)
        else:
            output = temporal_feature
            output = self.linear_head(output)

        
        output = rearrange(output, 'b n d -> b d n')
        output = self.cluster.revin(output, "denorm")
        tmp_future.append(output)
        tmp_future = torch.cat(tmp_future,2)
        output = self.final_layer(tmp_future)

        return output.reshape(BS,self.future_steps,self.n_vars,self.mul)






## Copyright https://github.com/thuml/iTransformer?tab=MIT-1-ov-file#readme
## Modified for notation alignmenet and batch structure
## extended to what inside itransformer folder

import torch
import torch.nn as nn
import numpy as np
from .itransformer.Transformer_EncDec import Encoder, EncoderLayer
from .itransformer.SelfAttention_Family import FullAttention, AttentionLayer
from .itransformer.Embed import DataEmbedding_inverted
from ..data_structure.utils import beauty_string
from .utils import  get_scope,get_activation,Embedding_cat_variables

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base



class ITransformer(Base):
    handle_multivariate = True
    handle_future_covariates = True
    handle_categorical_variables = True
    handle_quantile_loss = True
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    def __init__(self, 
                 # specific params
                 hidden_size:int,
                 d_model: int,
                 n_head: int,
                 n_layer_decoder: int,
                 use_norm: bool,
                 class_strategy: str = 'projection', #projection/average/cls_token    
                 dropout_rate: float=0.1,
                 activation: str='',
                 **kwargs)->None:
        """Initialize the ITransformer model for time series forecasting.
        
        This class implements the Inverted Transformer architecture as described in the paper 
        "ITRANSFORMER: INVERTED TRANSFORMERS ARE EFFECTIVE FOR TIME SERIES FORECASTING" 
        (https://arxiv.org/pdf/2310.06625).
        
        Args:
            hidden_size (int): The first embedding size of the model ('r' in the paper).
            d_model (int): The second embedding size (r^{tilda} in the model). Should be smaller than hidden_size.
            n_head (int): The number of attention heads.
            n_layer_decoder (int): The number of layers in the decoder.
            use_norm (bool): Flag to indicate whether to use normalization.
            class_strategy (str, optional): The strategy for classification, can be 'projection', 'average', or 'cls_token'. Defaults to 'projection'.
            dropout_rate (float, optional): The dropout rate for regularization. Defaults to 0.1.
            activation (str, optional): The activation function to be used. Defaults to ''.
            **kwargs: Additional keyword arguments.
        
        Raises:
            ValueError: If the activation function is not recognized.
        """
        
        
        
        
        super().__init__(**kwargs)
        if activation == 'torch.nn.SELU':
            beauty_string('SELU do not require BN','info',self.verbose)
            use_bn = False
        if isinstance(activation,str):
            activation = get_activation(activation)
        self.save_hyperparameters(logger=False)

        self.emb_past = Embedding_cat_variables(self.past_steps,self.emb_dim,self.embs_past, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        self.emb_fut = Embedding_cat_variables(self.future_steps,self.emb_dim,self.embs_fut, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        emb_past_out_channel = self.emb_past.output_channels
        emb_fut_out_channel = self.emb_fut.output_channels


  
        self.output_attention = False## not need output attention
        self.use_norm = use_norm
        # Embedding
        self.enc_embedding = DataEmbedding_inverted(self.past_steps, d_model, embed_type='what?', freq='what?', dropout=dropout_rate)  ##embed, freq not used inside
        self.class_strategy = class_strategy
        # Encoder-only architecture
        self.encoder = Encoder(
            [
                EncoderLayer(
                    AttentionLayer(
                        FullAttention(False, factor=0.1, attention_dropout=dropout_rate, ##factor is not used in the Full attention
                                      output_attention=self.output_attention), d_model, n_head), ## not need output attention
                    d_model,
                    hidden_size,
                    dropout = dropout_rate,
                    activation = activation()
                ) for l in range(n_layer_decoder)
            ],
            norm_layer=torch.nn.LayerNorm(d_model)
        )
        self.projector = nn.Linear(d_model, self.future_steps*self.mul, bias=True)

    def can_be_compiled(self):
        return True  

    def forecast(self, x_enc, x_mark_enc, x_dec, x_mark_dec):
        if self.use_norm:
            # Normalization from Non-stationary Transformer
            means = x_enc.mean(1, keepdim=True).detach()
            x_enc = x_enc - means
            stdev = torch.sqrt(torch.var(x_enc, dim=1, keepdim=True, unbiased=False) + 1e-5)
            x_enc /= stdev

        _, _, N = x_enc.shape # B L N
        # B: batch_size;    E: d_model; 
        # L: seq_len;       S: pred_len;
        # N: number of variate (tokens), can also includes covariates

        # Embedding
        # B L N -> B N E                (B L N -> B L E in the vanilla Transformer)
        enc_out = self.enc_embedding(x_enc, x_mark_enc) # covariates (e.g timestamp) can be also embedded as tokens
        
        # B N E -> B N E                (B L E -> B L E in the vanilla Transformer)
        # the dimensions of embedded time series has been inverted, and then processed by native attn, layernorm and ffn modules
        enc_out, attns = self.encoder(enc_out, attn_mask=None)

        # B N E -> B N S -> B S N 
        dec_out = self.projector(enc_out).permute(0, 2, 1)[:, :, :N] # filter the covariates

        if self.use_norm:

            # De-Normalization from Non-stationary Transformer
            dec_out = dec_out * (stdev[:, 0, :].unsqueeze(1).repeat(1, self.future_steps*self.mul, 1))
            dec_out = dec_out + (means[:, 0, :].unsqueeze(1).repeat(1, self.future_steps*self.mul, 1))


        return dec_out

    def forward(self, batch:dict)-> float:

        x_enc = batch['x_num_past'].to(self.device)
        BS = x_enc.shape[0]
        if 'x_cat_future' in batch.keys():
            emb_fut = self.emb_fut(BS,batch['x_cat_future'].to(self.device))
        else:
            emb_fut = self.emb_fut(BS,None)
        if 'x_cat_past' in batch.keys():
            emb_past = self.emb_past(BS,batch['x_cat_past'].to(self.device))
        else:
            emb_past = self.emb_past(BS,None)
 



        ##row 124 Transformer/experiments/exp_long_term_forecasting.py ma in realta' NON USATO!
        x_dec = torch.zeros(x_enc.shape[0],self.past_steps,self.out_channels).float().to(self.device)
        x_dec = torch.cat([batch['y'].to(self.device), x_dec], dim=1).float()

        dec_out = self.forecast(x_enc, emb_past, x_dec, emb_fut)
        idx_target = batch['idx_target'][0]
        return dec_out[:, :,idx_target].reshape(BS,self.future_steps,self.out_channels,self.mul)
        
        #return dec_out[:, -self.pred_len:, :]  # [B, L, D]



import torch
import torch.nn as nn
import numpy as np
from .tft import sub_nn

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from .utils import  QuantileLossMO
from typing import List, Union
from ..data_structure.utils import beauty_string
from .utils import  get_scope
from .utils import Embedding_cat_variables


class TIDE(Base):
    handle_multivariate = True
    handle_future_covariates = True
    handle_categorical_variables = True
    handle_quantile_loss = True
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    def __init__(self, 
                 hidden_size:int,
                 d_model: int,
                 n_add_enc: int,
                 n_add_dec: int,
                 dropout_rate: float,
                 activation: str='',
                 **kwargs)->None:
        """Initializes the model with specified parameters for a neural network architecture. Long-term Forecasting with TiDE: Time-series Dense Encoder
        https://arxiv.org/abs/2304.08424
        
        Args:
            hidden_size (int): The size of the hidden layers.
            d_model (int): The dimensionality of the model.
            n_add_enc (int): The number of additional encoder layers.
            n_add_dec (int): The number of additional decoder layers.
            dropout_rate (float): The dropout rate to be applied in the layers.
            activation (str, optional): The activation function to be used. Defaults to an empty string.
            **kwargs: Additional keyword arguments passed to the parent class.
    
        """
        
        
        
        super().__init__(**kwargs)
        self.save_hyperparameters(logger=False)

        # self.dropout = dropout_rate
   

        
        self.hidden_size = hidden_size # r
        self.d_model = d_model # r^tilda


        # for other numerical variables in the past
        self.aux_past_channels = self.past_channels - self.out_channels
        self.linear_aux_past = nn.ModuleList([nn.Linear(1, self.hidden_size) for _ in range(self.aux_past_channels)])

        # for numerical variables in the future
        self.aux_fut_channels = self.future_channels
        self.linear_aux_fut = nn.ModuleList([nn.Linear(1, self.hidden_size) for _ in range(self.aux_fut_channels)])
        
        #changinv from 1.1.5
        # embedding categorical for both past and future (ASSUMING BOTH AVAILABLE OR NO ONE)
        #self.seq_len = self.past_steps + self.future_steps
        #self.emb_cat_var = sub_nn.embedding_cat_variables(self.seq_len, future_steps, hidden_size, embs, self.device)

        self.emb_past = Embedding_cat_variables(self.past_steps,self.emb_dim,self.embs_past, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        self.emb_fut = Embedding_cat_variables(self.future_steps,self.emb_dim,self.embs_fut, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        emb_past_out_channel = self.emb_past.output_channels
        emb_fut_out_channel = self.emb_fut.output_channels


        ## FEATURE PROJECTION
        # past
        if self.aux_past_channels>0:
            self.feat_proj_past = ResidualBlock(hidden_size+emb_past_out_channel, d_model, dropout_rate, activation)
        else:
            self.feat_proj_past = ResidualBlock(emb_past_out_channel, d_model, dropout_rate, activation)
        # future
        if self.aux_fut_channels>0:
            self.feat_proj_fut = ResidualBlock(hidden_size+emb_fut_out_channel, d_model, dropout_rate, activation)
        else:
            self.feat_proj_fut = ResidualBlock(emb_fut_out_channel, d_model, dropout_rate, activation)

        # # ENCODER
        self.enc_dim_input = self.past_steps*self.out_channels + (self.past_steps+self.future_steps)*d_model 
        self.enc_dim_output = self.future_steps*d_model
        self.first_encoder = ResidualBlock(self.enc_dim_input, self.enc_dim_output, dropout_rate, activation)
        self.aux_encoder = nn.ModuleList([ResidualBlock(self.enc_dim_output, self.enc_dim_output, dropout_rate, activation) for _ in range(1, n_add_enc)])

        # # DECODER
        self.first_decoder = ResidualBlock(self.enc_dim_output, self.enc_dim_output, dropout_rate, activation)
        self.aux_decoder = nn.ModuleList([ResidualBlock(self.enc_dim_output, self.enc_dim_output, dropout_rate, activation) for _ in range(1, n_add_dec)])

        ## TEMPORAL DECOER
        self.temporal_decoder = ResidualBlock(2*d_model, self.out_channels*self.mul, dropout_rate, activation)

        # linear for Y lookback
        self.linear_target = nn.Linear(self.past_steps*self.out_channels, self.future_steps*self.out_channels*self.mul)
    
    def can_be_compiled(self):
        return False  
  

    def forward(self, batch:dict)-> float:
        """training process of the diffusion network

        Args:
            batch (dict): variables loaded

        Returns:
            float: total loss about the prediction of the noises over all subnets extracted
        """

        # LOADING AUTOREGRESSIVE CONTEXT OF TARGET VARIABLES

        num_past = batch['x_num_past'].to(self.device)
        idx_target = batch['idx_target'][0]
        y_past = num_past[:,:,idx_target]
        B = y_past.shape[0]

        # LOADING EMBEDDING CATEGORICAL VARIABLES
        #emb_cat_past, emb_cat_fut = self.cat_categorical_vars(batch)
    
        if 'x_cat_future' in batch.keys():
            emb_fut = self.emb_fut(B,batch['x_cat_future'].to(self.device))
        else:
            emb_fut = self.emb_fut(B,None)
        if 'x_cat_past' in batch.keys():
            emb_past = self.emb_past(B,batch['x_cat_past'].to(self.device))
        else:
            emb_past = self.emb_past(B,None)


      
        #emb_cat_past = torch.mean(emb_cat_past, dim = 2)
        #emb_cat_fut = torch.mean(emb_cat_fut, dim = 2)
      
        ### LOADING PAST AND FUTURE NUMERICAL VARIABLES
        # load in the model auxiliar numerical variables

        if self.aux_past_channels>0: # if we have more numerical variables about past
            aux_num_past = self.remove_var(num_past, idx_target, 2) # remove the autoregressive variable
            assert self.aux_past_channels == aux_num_past.size(2),  beauty_string(f"{self.aux_past_channels} LAYERS FOR PAST VARS AND {aux_num_past.size(2)} VARS",'section',True) # to check if we are using the expected number of variables about past
            # concat all embedded vars and mean of them
            aux_emb_num_past = torch.Tensor().to(self.device)
            for i, layer in enumerate(self.linear_aux_past):
                aux_emb_past = layer(aux_num_past[:,:,[i]]).unsqueeze(2)
                aux_emb_num_past = torch.cat((aux_emb_num_past, aux_emb_past), dim=2)
            aux_emb_num_past = torch.mean(aux_emb_num_past, dim = 2)
        else: 
            aux_emb_num_past = None # non available vars
            
        if self.aux_fut_channels>0: # if we have more numerical variables about future
            # AUX means AUXILIARY variables
            aux_num_fut = batch['x_num_future'].to(self.device)
            assert self.aux_fut_channels == aux_num_fut.size(2), beauty_string(f"{self.aux_fut_channels} LAYERS FOR PAST VARS AND {aux_num_fut.size(2)} VARS",'section',True)  # to check if we are using the expected number of variables about fut
            # concat all embedded vars and mean of them
            aux_emb_num_fut = torch.Tensor().to(self.device)
            for j, layer in enumerate(self.linear_aux_fut):
                aux_emb_fut = layer(aux_num_fut[:,:,[j]]).unsqueeze(2)
                aux_emb_num_fut = torch.cat((aux_emb_num_fut, aux_emb_fut), dim=2)
            aux_emb_num_fut = torch.mean(aux_emb_num_fut, dim = 2)
        else:
            aux_emb_num_fut = None # non available vars

        # past^tilda

        if self.aux_past_channels>0:
            emb_past = torch.cat((emb_past, aux_emb_num_past), dim = 2) # [B, L, 2R] #
            proj_past = self.feat_proj_past(emb_past, True) # [B, L, R^tilda] #
        else:
            proj_past = self.feat_proj_past(emb_past, True) # [B, L, R^tilda] #

        # fut^tilda
        if self.aux_fut_channels>0:
            emb_fut = torch.cat((emb_fut, aux_emb_num_fut), dim = 2) # [B, H, 2R] #
            proj_fut = self.feat_proj_fut(emb_fut, True) # [B, H, R^tilda] #
        else:
            proj_fut = self.feat_proj_fut(emb_fut, True) # [B, H, R^tilda] #
            
        concat = torch.cat((y_past.view(B, -1), proj_past.view(B, -1), proj_fut.view(B, -1)), dim = 1) # [B, L*self.mul + (L+H)*R^tilda] #
        dense_enc = self.first_encoder(concat)
        for lay_enc in self.aux_encoder:
            dense_enc = lay_enc(dense_enc)

        dense_dec = self.first_decoder(dense_enc)
        for lay_dec in self.aux_decoder:
            dense_dec = lay_dec(dense_dec)
        
        temp_dec_input = torch.cat((dense_dec.view(B, self.future_steps, self.d_model), proj_fut), dim = 2)
        temp_dec_output = self.temporal_decoder(temp_dec_input, False)
        temp_dec_output = temp_dec_output.view(B, self.future_steps, self.out_channels, self.mul)

        linear_regr = self.linear_target(y_past.view(B, -1))
        linear_output = linear_regr.view(B, self.future_steps, self.out_channels, self.mul)

        output = temp_dec_output + linear_output
        return output

    '''
    # function to concat embedded categorical variables
    def cat_categorical_vars(self, batch:dict):
        """Extracting categorical context about past and future

        Args:
            batch (dict): Keys checked -> ['x_cat_past', 'x_cat_future']

        Returns:
            List[torch.Tensor, torch.Tensor]: cat_emb_past, cat_emb_fut
        """
        cat_past = None
        cat_fut = None
        # GET AVAILABLE CATEGORICAL CONTEXT
        if 'x_cat_past' in batch.keys():
            cat_past = batch['x_cat_past'].to(self.device)
        if 'x_cat_future' in batch.keys():
            cat_fut = batch['x_cat_future'].to(self.device)
        # CONCAT THEM, according to self.emb_cat_var usage  
        if cat_past is None:
            emb_cat_full = self.emb_cat_var(batch['x_num_past'].shape[0],self.device)

        else:
            cat_full = torch.cat((cat_past, cat_fut), dim = 1)
            emb_cat_full = self.emb_cat_var(cat_full,self.device)
        cat_emb_past = emb_cat_full[:,:self.past_steps,:,:]
        cat_emb_fut = emb_cat_full[:,-self.future_steps:,:,:]

        return cat_emb_past, cat_emb_fut

    #function to extract from batch['x_num_past'] all variables except the one autoregressive
    '''
    def remove_var(self, tensor: torch.Tensor, indexes_to_exclude: list, dimension: int)-> torch.Tensor:
        """Function to remove variables from tensors in chosen dimension and position 

        Args:
            tensor (torch.Tensor): starting tensor
            indexes_to_exclude (list): index of the chosen dimension we want t oexclude
            dimension (int): dimension of the tensor on which we want to work (not list od dims!!)

        Returns:
            torch.Tensor: new tensor without the chosen variables
        """

        remaining_idx = torch.tensor([i for i in range(tensor.size(dimension)) if i not in indexes_to_exclude]).to(tensor.device)
        # Select the desired sub-tensor
        extracted_subtensors = torch.index_select(tensor, dim=dimension, index=remaining_idx)
        
        return extracted_subtensors


class ResidualBlock(nn.Module):
    def __init__(self, in_size:int, out_size:int, dropout_rate:float, activation_fun:str=''):
        """Residual Block as basic layer of the archetecture. 

        MLP with one hidden layer, activation and skip connection
        Basically dimension d_model, but better if input_dim and output_dim are explicit

        in_size and out_size to handle dimensions at different stages of the NN

        Args:
            in_size (int): 
            out_size (int): 
            dropout_rate (float): 
            activation_fun (str, optional): activation function to use in the Residual Block. Defaults to nn.ReLU.
        """
        super().__init__()

        self.direct_linear = nn.Linear(in_size, out_size, bias = False)

        if activation_fun=='':
            self.act = nn.ReLU()
        else:
            activation = eval(activation_fun)
            self.act = activation()
        self.lin = nn.Linear(in_size, out_size)
        self.dropout = nn.Dropout(dropout_rate)

        self.final_norm = nn.LayerNorm(out_size)


    def forward(self, x, apply_final_norm = True):
        direct_x = self.direct_linear(x)

        x = self.dropout(self.lin(self.act(x)))

        out = x + direct_x
        if apply_final_norm:
            return self.final_norm(out)
        return out


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.fft


class Inception_Block_V1(nn.Module):
    def __init__(self, in_channels, out_channels, num_kernels=6, init_weight=True):
        super(Inception_Block_V1, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_kernels = num_kernels
        kernels = []
        for i in range(self.num_kernels):
            kernels.append(nn.Conv2d(in_channels, out_channels, kernel_size=2 * i + 1, padding=i))
        self.kernels = nn.ModuleList(kernels)
        if init_weight:
            self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward(self, x):
        res_list = []
        for i in range(self.num_kernels):
            res_list.append(self.kernels[i](x))
        res = torch.stack(res_list, dim=-1).mean(-1)
        return res

def FFT_for_Period(x, k=2):
    # [B, T, C]
    xf = torch.fft.rfft(x, dim=1)
    # find period by amplitudes
    frequency_list = abs(xf).mean(0).mean(-1)
    frequency_list[0] = 0
    _, top_list = torch.topk(frequency_list, k)
    top_list = top_list.detach().cpu().numpy()
    period = x.shape[1] // top_list
    return period, abs(xf).mean(-1)[:, top_list]



class TimesBlock(nn.Module):
    def __init__(self, seq_len,pred_len,top_k,d_model,d_ff,num_kernels):
        super(TimesBlock, self).__init__()
        self.seq_len = seq_len
        self.pred_len = pred_len
        self.k = top_k
        # parameter-efficient design
        self.conv = nn.Sequential(
            Inception_Block_V1(d_model, d_ff,
                               num_kernels=num_kernels),
            nn.GELU(),
            Inception_Block_V1(d_ff, d_model,
                               num_kernels=num_kernels)
        )

    def forward(self, x):
        B, T, N = x.size()

        period_list, period_weight = FFT_for_Period(x, self.k)

        res = []
        for i in range(self.k):
            period = period_list[i]
            # padding
            if (self.seq_len + self.pred_len) % period != 0:
                length = (
                                 ((self.seq_len + self.pred_len) // period) + 1) * period
                padding = torch.zeros([x.shape[0], (length - (self.seq_len + self.pred_len)), x.shape[2]]).to(x.device)
                out = torch.cat([x, padding], dim=1)
            else:
                length = (self.seq_len + self.pred_len)
                out = x
            # reshape
            out = out.reshape(B, length // period, period,
                              N).permute(0, 3, 1, 2).contiguous()
            # 2D conv: from 1d Variation to 2d Variation
            out = self.conv(out)
            # reshape back
            out = out.permute(0, 2, 3, 1).reshape(B, -1, N)
            res.append(out[:, :(self.seq_len + self.pred_len), :])
        res = torch.stack(res, dim=-1)
        # adaptive aggregation
        period_weight = F.softmax(period_weight, dim=1)
        period_weight = period_weight.unsqueeze(
            1).unsqueeze(1).repeat(1, T, N, 1)
        res = torch.sum(res * period_weight, -1)
        # residual connection
        res = res + x
        return res



# Default
import os
import yaml
import enum

import numpy as np
import pandas as pd
from importlib import resources
from typing import Optional, Union
from pandas.tseries.frequencies import to_offset

# Custom
from .configuration_tinytimemixer import TinyTimeMixerConfig
from .modeling_tinytimemixer import TinyTimeMixerForPrediction
from .consts import DEFAULT_FREQUENCY_MAPPING, TTM_LOW_RESOLUTION_MODELS_MAX_CONTEXT

# Hugging face
from transformers import PreTrainedModel

# PyTorch
import torch

##TODO fix
TTM_CONF = {'ibm-granite-models': {'512-96-r1': {'release': 'r1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r1', 'revision': 'main', 'context_length': 512, 'prediction_length': 96}, '1024-96-r1': {'release': 'r1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r1', 'revision': '1024_96_v1', 'context_length': 1024, 'prediction_length': 96}, '512-96-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': 'main', 'context_length': 512, 'prediction_length': 96}, '512-192-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '512-192-r2', 'context_length': 512, 'prediction_length': 192}, '512-336-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '512-336-r2', 'context_length': 512, 'prediction_length': 336}, '512-720-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '512-720-r2', 'context_length': 512, 'prediction_length': 720}, '1024-96-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '1024-96-r2', 'context_length': 1024, 'prediction_length': 96}, '1024-192-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '1024-192-r2', 'context_length': 1024, 'prediction_length': 192}, '1024-336-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '1024-336-r2', 'context_length': 1024, 'prediction_length': 336}, '1024-720-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '1024-720-r2', 'context_length': 1024, 'prediction_length': 720}, '1536-96-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '1536-96-r2', 'context_length': 1536, 'prediction_length': 96}, '1536-192-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '1536-192-r2', 'context_length': 1536, 'prediction_length': 192}, '1536-336-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '1536-336-r2', 'context_length': 1536, 'prediction_length': 336}, '1536-720-r2': {'release': 'r2', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '1536-720-r2', 'context_length': 1536, 'prediction_length': 720}, '52-16-ft-r2.1': {'release': 'r2.1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '52-16-ft-r2.1', 'context_length': 52, 'prediction_length': 16}, '52-16-ft-l1-r2.1': {'release': 'r2.1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '52-16-ft-l1-r2.1', 'context_length': 52, 'prediction_length': 16}, '90-30-ft-r2.1': {'release': 'r2.1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '90-30-ft-r2.1', 'context_length': 90, 'prediction_length': 30}, '90-30-ft-l1-r2.1': {'release': 'r2.1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '90-30-ft-l1-r2.1', 'context_length': 90, 'prediction_length': 30}, '180-60-ft-l1-r2.1': {'release': 'r2.1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '180-60-ft-l1-r2.1', 'context_length': 180, 'prediction_length': 60}, '360-60-ft-l1-r2.1': {'release': 'r2.1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '360-60-ft-l1-r2.1', 'context_length': 360, 'prediction_length': 60}, '512-48-ft-r2.1': {'release': 'r2.1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '512-48-ft-r2.1', 'context_length': 512, 'prediction_length': 48}, '512-48-ft-l1-r2.1': {'release': 'r2.1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '512-48-ft-l1-r2.1', 'context_length': 512, 'prediction_length': 48}, '512-96-ft-r2.1': {'release': 'r2.1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '512-96-ft-r2.1', 'context_length': 512, 'prediction_length': 96}, '512-96-ft-l1-r2.1': {'release': 'r2.1', 'model_card': 'ibm-granite/granite-timeseries-ttm-r2', 'revision': '512-96-ft-l1-r2.1', 'context_length': 512, 'prediction_length': 96}}, 'research-use-models': {'512-96-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': 'main', 'context_length': 512, 'prediction_length': 96}, '512-192-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '512-192-ft-r2', 'context_length': 512, 'prediction_length': 192}, '512-336-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '512-336-ft-r2', 'context_length': 512, 'prediction_length': 336}, '512-720-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '512-720-ft-r2', 'context_length': 512, 'prediction_length': 720}, '1024-96-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '1024-96-ft-r2', 'context_length': 1024, 'prediction_length': 96}, '1024-192-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '1024-192-ft-r2', 'context_length': 1024, 'prediction_length': 192}, '1024-336-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '1024-336-ft-r2', 'context_length': 1024, 'prediction_length': 336}, '1024-720-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '1024-720-ft-r2', 'context_length': 1024, 'prediction_length': 720}, '1536-96-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '1536-96-ft-r2', 'context_length': 1536, 'prediction_length': 96}, '1536-192-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '1536-192-ft-r2', 'context_length': 1536, 'prediction_length': 192}, '1536-336-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '1536-336-ft-r2', 'context_length': 1536, 'prediction_length': 336}, '1536-720-ft-r2': {'release': 'r2', 'model_card': 'ibm-research/ttm-research-r2', 'revision': '1536-720-ft-r2', 'context_length': 1536, 'prediction_length': 720}}}


class ForceReturn(enum.Enum):
    """`Enum` for the `force_return` parameter in the `get_model` function.

    "zeropad" = Returns a pre-trained TTM that has a context length higher than the input context length, hence,
        the user must apply zero-padding to use the returned model.
    "rolling" = Returns a pre-trained TTM that has a prediction length lower than the requested prediction length,
        hence, the user must apply rolling technique to use the returned model to forecast to the desired length.
        The `RecursivePredictor` class can be utilized in this scenario.
    "random_init_small" = Returns a randomly initialized small TTM which must be trained before performing inference.
    "random_init_medium" = Returns a randomly initialized medium TTM which must be trained before performing inference.
    "random_init_large" = Returns a randomly initialized large TTM which must be trained before performing inference.

    """

    ZEROPAD = "zeropad"
    ROLLING = "rolling"
    RANDOM_INIT_SMALL = "random_init_small"
    RANDOM_INIT_MEDIUM = "random_init_medium"
    RANDOM_INIT_LARGE = "random_init_large"

class ModelSize(enum.Enum):
    """`Enum` for the `size` parameter in the `get_random_ttm` function."""

    SMALL = "small"
    MEDIUM = "medium"
    LARGE = "large"


def check_ttm_model_path(model_path):
    if (
        "ibm/TTM" in model_path
        or "ibm-granite/granite-timeseries-ttm-r1" in model_path
        or "ibm-granite/granite-timeseries-ttm-v1" in model_path
        or "ibm-granite/granite-timeseries-ttm-1m" in model_path
    ):
        return 1
    elif "ibm-granite/granite-timeseries-ttm-r2" in model_path:
        return 2
    elif "ibm-research/ttm-research-r2" in model_path:
        return 3
    else:
        return 0
    
def count_parameters(model: torch.nn.Module) -> int:
    """Count trainable parameters in a model

    Args:
        model (torch.nn.Module): The model.

    Returns:
        int: Number of parameters requiring gradients.
    """
    return sum(p.numel() for p in model.parameters() if p.requires_grad)
    
def get_random_ttm(
    context_length: int, prediction_length: int, size: str = ModelSize.SMALL.value, **kwargs
) -> PreTrainedModel:
    """Get a TTM with random weights.

    Args:
        context_length (int): Context length or history.
        prediction_length (int): Prediction length or forecast horizon.
        size (str, optional): Size of the desired TTM (small/medium/large). Defaults to "small".

    Raises:
        ValueError: If wrong size is provided.
        ValueError: Context length should be at least 4 if `size=small`,
            or at least 16 if `size=medium`,
            or at least 32 if `size=large`.

    Returns:
        PreTrainedModel: TTM model with randomly initialized weights.
    """
    if ModelSize.SMALL.value in size.lower():
        cl_lower_bound = 4
        apl = 0
    elif ModelSize.MEDIUM.value in size.lower():
        cl_lower_bound = 16
        apl = 3
    elif ModelSize.LARGE.value in size.lower():
        cl_lower_bound = 32
        apl = 5
    else:
        raise ValueError("Wrong size. Should be either of these [small/medium/large].")
    if context_length < cl_lower_bound:
        raise ValueError(f"Context length should be at least {cl_lower_bound} if `size={size}`.")

    cl = context_length if context_length % 2 == 0 else context_length - 1

    pl = 2
    while cl % pl == 0 and cl / pl >= 8:
        pl = pl * 2

    if ModelSize.SMALL.value in size.lower():
        d_model = 2 * pl
        num_layers = 3
    elif ModelSize.MEDIUM.value in size.lower():
        d_model = 16 * 2**apl
        num_layers = 3
    elif ModelSize.LARGE.value in size.lower():
        d_model = 16 * 2**apl
        num_layers = 5
    else:
        raise ValueError("Wrong size. Should be either of these [small/medium/large].")

    ttm_config = TinyTimeMixerConfig(
        context_length=cl,
        prediction_length=prediction_length,
        patch_length=pl,
        patch_stride=pl,
        d_model=d_model,
        num_layers=num_layers,
        decoder_num_layers=2,
        decoder_d_model=d_model,
        adaptive_patching_levels=apl,
        dropout=0.2,
        **kwargs,
    )
    model = TinyTimeMixerForPrediction(config=ttm_config)

    return model

def get_frequency_token(token_name: str):
        token = DEFAULT_FREQUENCY_MAPPING.get(token_name, None)
        if token is not None:
            return torch.tensor(token, dtype=torch.int)

        # try to map as a frequency string
        try:
            token_name_offs = to_offset(token_name).freqstr
            token = DEFAULT_FREQUENCY_MAPPING.get(token_name_offs, None)
            if token is not None:
                return torch.tensor(token, dtype=torch.int)
        except ValueError:
            # lastly try to map the timedelta to a frequency string
            token_name_td = pd._libs.tslibs.timedeltas.Timedelta(token_name)
            token_name_offs = to_offset(token_name_td).freqstr
            token = DEFAULT_FREQUENCY_MAPPING.get(token_name_offs, None)
            if token is not None:
                return torch.tensor(token, dtype=torch.int)
            
        token = DEFAULT_FREQUENCY_MAPPING["oov"]

        return torch.tensor(token, dtype=torch.int)

    
class RMSELoss(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.mse = torch.nn.MSELoss()
        
    def forward(self,yhat,y):
        return torch.sqrt(self.mse(yhat,y))

def get_model(
    model_path: str,
    model_name: str = "ttm",
    context_length: Optional[int] = None,
    prediction_length: Optional[int] = None,
    freq_prefix_tuning: bool = False,
    freq: Optional[str] = None,
    prefer_l1_loss: bool = False,
    prefer_longer_context: bool = True,
    force_return: Optional[str] = None,
    return_model_key: bool = False,
    **kwargs,
) -> Union[str, PreTrainedModel]:
    """TTM Model card offers a suite of models with varying `context_length` and `prediction_length` combinations.
    This wrapper automatically selects the right model based on the given input `context_length` and
    `prediction_length` abstracting away the internal complexity.

    Args:
        model_path (str): HuggingFace model card path or local model path (Ex. ibm-granite/granite-timeseries-ttm-r2)
        model_name (str, optional): Model name to use. Current allowed values: [ttm]. Defaults to "ttm".
        context_length (int, optional): Input Context length or history. Defaults to None.
        prediction_length (int, optional): Length of the forecast horizon. Defaults to None.
        freq_prefix_tuning (bool, optional): If true, it will prefer TTM models that are trained with frequency prefix
            tuning configuration. Defaults to None.
        freq (str, optional): Resolution or frequency of the data. Defaults to None. Allowed values are as per the
            `tsfm_public.toolkit.time_series_preprocessor.DEFAULT_FREQUENCY_MAPPING`.
            See this for details: https://github.com/ibm-granite/granite-tsfm/blob/main/tsfm_public/toolkit/time_series_preprocessor.py.
        prefer_l1_loss (bool, optional): If True, it will prefer choosing models that were trained with L1 loss or
            mean absolute error loss. Defaults to False.
        prefer_longer_context (bool, optional): If True, it will prefer selecting model with longer context/history
            Defaults to True.
        force_return (str, optional): This is used to force the get_model() to return a TTM model even when the provided
            configurations don't match with the existing TTMs. It gets the closest TTM possible. Allowed values are
            ["zeropad"/"rolling"/"random_init_small"/"random_init_medium"/"random_init_large"/`None`].
            "zeropad" = Returns a pre-trained TTM that has a context length higher than the input context length, hence,
            the user must apply zero-padding to use the returned model.
            "rolling" = Returns a pre-trained TTM that has a prediction length lower than the requested prediction length,
            hence, the user must apply rolling technique to use the returned model to forecast to the desired length.
            The `RecursivePredictor` class can be utilized in this scenario.
            "random_init_small" = Returns a randomly initialized small TTM which must be trained before performing inference.
            "random_init_medium" = Returns a randomly initialized medium TTM which must be trained before performing inference.
            "random_init_large" = Returns a randomly initialized large TTM which must be trained before performing inference.
            `None` = `force_return` is disable. Raises an error if no suitable model is found.
            Defaults to None.
        return_model_key (bool, optional): If True, only the TTM model name will be returned, instead of the actual model.
            This does not downlaod the model, and only returns the name of the suitable model. Defaults to False.

    Returns:
        Union[str, PreTrainedModel]: Returns the Model, or the model name.
    """
    if model_name.lower() == "ttm":
        model_path_type = check_ttm_model_path(model_path)
        prediction_filter_length = None
        ttm_model_revision = None
        if model_path_type != 0:
            if context_length is None or prediction_length is None:
                raise ValueError(
                    "Provide `context_length` and `prediction_length` when `model_path` is a hugginface model path."
                )

            # Get freq
            R = DEFAULT_FREQUENCY_MAPPING.get(freq, 0)

            # Get list of all TTM models
            '''
            config_dir = resources.files("tsfm_public.resources.model_paths_config")
            with open(os.path.join(config_dir, "ttm.yaml"), "r") as file:
                model_revisions = yaml.safe_load(file)
            '''
            model_revisions = TTM_CONF ##TODO fix this
            if model_path_type == 1 or model_path_type == 2:
                available_models = model_revisions["ibm-granite-models"]
                filtered_models = {}
                if model_path_type == 1:
                    for k in available_models.keys():
                        if available_models[k]["release"].startswith("r1"):
                            filtered_models[k] = available_models[k]
                if model_path_type == 2:
                    for k in available_models.keys():
                        if available_models[k]["release"].startswith("r2"):
                            filtered_models[k] = available_models[k]
                available_models = filtered_models
            else:
                available_models = model_revisions["research-use-models"]

            # Calculate shortest TTM context length, will be needed later
            available_model_keys = list(available_models.keys())
            available_ttm_context_lengths = [available_models[m]["context_length"] for m in available_model_keys]
            shortest_ttm_context_length = min(available_ttm_context_lengths)

            # Step 1: Filter models based on freq (R)
            if model_path_type == 1 or model_path_type == 2:
                # Only, r2.1 models are suitable for Daily or longer freq
                if R >= 8:
                    models = [m for m in available_models.keys() if "r2.1" in available_models[m]["release"]]
                else:
                    models = list(available_models.keys())
            else:
                models = list(available_models.keys())

            # Step 2: Filter models by context length constraint
            # Choose all models which have lower context length than
            # the input available length
            selected_models_ = []
            if context_length < shortest_ttm_context_length:
                if force_return is None:
                    raise ValueError(
                        "Requested context length is less than the "
                        f"shortest context length for TTMs: {shortest_ttm_context_length}. "
                        "Set `force_return=zeropad` to get a TTM with longer context."
                    )
                elif force_return == ForceReturn.ZEROPAD.value:  # force_return.startswith("zero"):
                    # Keep all models. Zero-padding must be done outside.
                    selected_models_ = models
            else:
                lowest_context_length = np.inf
                shortest_context_models = []
                for m in models:
                    if available_models[m]["context_length"] <= context_length:
                        selected_models_.append(m)
                    if available_models[m]["context_length"] <= lowest_context_length:
                        lowest_context_length = available_models[m]["context_length"]
                        shortest_context_models.append(m)

            if len(selected_models_) == 0:
                if force_return is None:
                    raise ValueError(
                        "Could not find a TTM with `context_length` shorter "
                        f"than the requested context length = {context_length}. "
                        "Set `force_return=zeropad` to get a TTM with longer context."
                    )
                elif force_return == ForceReturn.ZEROPAD.value:  # force_return.startswith("zero"):
                    selected_models_ = shortest_context_models
            models = selected_models_

            # Step 3: Apply L1 and FT preferences only when context_length <= 512
            if len(models) > 0:
                if prefer_longer_context:
                    reference_context = min(
                        context_length, max([available_models[m]["context_length"] for m in models])
                    )
                else:
                    reference_context = min([available_models[m]["context_length"] for m in models])
                if reference_context <= TTM_LOW_RESOLUTION_MODELS_MAX_CONTEXT:
                    # Step 3a: Filter based on L1 preference
                    if prefer_l1_loss:
                        l1_models = [m for m in models if "-l1-" in m]
                        if l1_models:
                            models = l1_models

                    # Step 3b: Filter based on frequency tuning indicator preference
                    if freq_prefix_tuning:
                        ft_models = [m for m in models if "-ft-" in m]
                        if ft_models:
                            models = ft_models

            # Step 4: Sort models by context length (descending if prefer_longer_context else ascending)
            # Step 5: Sub-sort for each context length by forecast length in ascending order
            if len(models) > 0:
                sign = -1 if prefer_longer_context else 1
                models = sorted(
                    models,
                    key=lambda m: (
                        sign * int(available_models[m]["context_length"]),
                        int(available_models[m]["prediction_length"]),
                    ),
                )

            # Step 6: Remove models whose forecast length is less than input forecast length
            # Because, this needs recursion which has to be handled outside this get_model() utility
            if len(models) > 0:
                selected_models_ = []
                highest_prediction_length = -np.inf
                highest_prediction_model = None
                for m in models:
                    if int(available_models[m]["prediction_length"]) >= prediction_length:
                        selected_models_.append(m)
                    if available_models[m]["prediction_length"] > highest_prediction_length:
                        highest_prediction_length = available_models[m]["prediction_length"]
                        highest_prediction_model = m
                if len(selected_models_) == 0:
                    if force_return is None:
                        raise ValueError(
                            "Could not find a TTM with `prediction_length` higher "
                            f"than the requested prediction length = {prediction_length}. "
                            "Set `force_return=rolling` to get a TTM with shorted prediction "
                            "length. Rolling must be done outside."
                        )
                    elif force_return == ForceReturn.ROLLING.value:  # force_return.startswith("roll"):
                        selected_models_.append(highest_prediction_model)
                models = selected_models_

            # Step 7: Do not allow unknow frequency
            if freq_prefix_tuning and (freq is not None) and (freq not in DEFAULT_FREQUENCY_MAPPING.keys()):
                models = []

            # Step 8: Return the first available model or a dummy model if none found
            if len(models) == 0:
                if force_return is None:
                    raise ValueError(
                        "No suitable pre-trained TTM was found! Set `force_return` to "
                        "random_init_small/random_init_medium/random_init_large "
                        "to get a randomly initialized TTM of size small/medium/large "
                        "respectively."
                    )
                elif force_return in [
                    ForceReturn.RANDOM_INIT_SMALL.value,
                    ForceReturn.RANDOM_INIT_MEDIUM.value,
                    ForceReturn.RANDOM_INIT_LARGE.value,
                ]:  # "sma" in force_return.lower() or "med" in force_return.lower() or "lar" in force_return.lower():
                    model = get_random_ttm(context_length, prediction_length, size=force_return)
                    if return_model_key:
                        model_key = force_return.split("_")[-1]
                        return f"TTM({model_key})"
                    else:
                        return model
                else:
                    raise ValueError(
                        "Could not find a suitable TTM for the given "
                        f"context_length = {context_length}, and "
                        f"prediction_length = {prediction_length}. "
                        "Check the model card for more information. "
                        "set `force_return` properly (see the docstrings) "
                        "if you want to get a randomly initialized TTM."
                    )
            else:
                model_key = models[0]

            # selected_context_length = available_models[model_key]["context_length"]
            selected_prediction_length = available_models[model_key]["prediction_length"]
            if selected_prediction_length > prediction_length:
                prediction_filter_length = prediction_length

            # if selected_prediction_length < prediction_length:
            #     LOGGER.warning(
            #         "Selected `prediction_length` is shorter than the requested "
            #         "length since no suitable model could be found. You can use "
            #         " `RecursivePredictor` for forecast to the desired length."
            #     )

            ttm_model_revision = available_models[model_key]["revision"]

        else:
            prediction_filter_length = prediction_length

        if return_model_key:
            return model_key
        # Load model
        model = TinyTimeMixerForPrediction.from_pretrained(
            model_path,
            revision=ttm_model_revision,
            prediction_filter_length=prediction_filter_length,
            **kwargs,
        )
    else:
        raise ValueError("Currently supported values for `model_name` = 'ttm'.")

    return model


import torch
import torch.nn as nn
from einops import rearrange, repeat


from math import sqrt

class FullAttention(nn.Module):
    '''
    The Attention operation
    '''
    def __init__(self, scale=None, attention_dropout=0.1):
        super(FullAttention, self).__init__()
        self.scale = scale
        self.dropout = nn.Dropout(attention_dropout)
        
    def forward(self, queries, keys, values):
        B, L, H, E = queries.shape
        _, S, _, D = values.shape
        scale = self.scale or 1./sqrt(E)

        scores = torch.einsum("blhe,bshe->bhls", queries, keys)
        A = self.dropout(torch.softmax(scale * scores, dim=-1))
        V = torch.einsum("bhls,bshd->blhd", A, values)
        
        return V.contiguous()


class AttentionLayer(nn.Module):
    '''
    The Multi-head Self-Attention (MSA) Layer
    '''
    def __init__(self, d_model, n_heads, d_keys=None, d_values=None, mix=True, dropout = 0.1):
        super(AttentionLayer, self).__init__()

        d_keys = d_keys or (d_model//n_heads)
        d_values = d_values or (d_model//n_heads)

        self.inner_attention = FullAttention(scale=None, attention_dropout = dropout)
        self.query_projection = nn.Linear(d_model, d_keys * n_heads)
        self.key_projection = nn.Linear(d_model, d_keys * n_heads)
        self.value_projection = nn.Linear(d_model, d_values * n_heads)
        self.out_projection = nn.Linear(d_values * n_heads, d_model)
        self.n_heads = n_heads
        self.mix = mix

    def forward(self, queries, keys, values):
        B, L, _ = queries.shape
        _, S, _ = keys.shape
        H = self.n_heads

        queries = self.query_projection(queries).view(B, L, H, -1)
        keys = self.key_projection(keys).view(B, S, H, -1)
        values = self.value_projection(values).view(B, S, H, -1)

        out = self.inner_attention(
            queries,
            keys,
            values,
        )
        if self.mix:
            out = out.transpose(2,1).contiguous()
        out = out.view(B, L, -1)

        return self.out_projection(out)

class TwoStageAttentionLayer(nn.Module):
    '''
    The Two Stage Attention (TSA) Layer
    input/output shape: [batch_size, Data_dim(D), Seg_num(L), d_model]
    '''
    def __init__(self, seg_num, factor, d_model, n_heads, d_ff = None, dropout=0.1):
        super(TwoStageAttentionLayer, self).__init__()
        d_ff = d_ff or 4*d_model
        self.time_attention = AttentionLayer(d_model, n_heads, dropout = dropout)
        self.dim_sender = AttentionLayer(d_model, n_heads, dropout = dropout)
        self.dim_receiver = AttentionLayer(d_model, n_heads, dropout = dropout)
        self.router = nn.Parameter(torch.randn(seg_num, factor, d_model))
        
        self.dropout = nn.Dropout(dropout)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.norm3 = nn.LayerNorm(d_model)
        self.norm4 = nn.LayerNorm(d_model)

        self.MLP1 = nn.Sequential(nn.Linear(d_model, d_ff),
                                nn.GELU(),
                                nn.Linear(d_ff, d_model))
        self.MLP2 = nn.Sequential(nn.Linear(d_model, d_ff),
                                nn.GELU(),
                                nn.Linear(d_ff, d_model))

    def forward(self, x):
        #Cross Time Stage: Directly apply MSA to each dimension
        batch = x.shape[0]
        time_in = rearrange(x, 'b ts_d seg_num d_model -> (b ts_d) seg_num d_model')
        time_enc = self.time_attention(
            time_in, time_in, time_in
        )
        dim_in = time_in + self.dropout(time_enc)
        dim_in = self.norm1(dim_in)
        dim_in = dim_in + self.dropout(self.MLP1(dim_in))
        dim_in = self.norm2(dim_in)

        #Cross Dimension Stage: use a small set of learnable vectors to aggregate and distribute messages to build the D-to-D connection
        dim_send = rearrange(dim_in, '(b ts_d) seg_num d_model -> (b seg_num) ts_d d_model', b = batch)
        batch_router = repeat(self.router, 'seg_num factor d_model -> (repeat seg_num) factor d_model', repeat = batch)
        dim_buffer = self.dim_sender(batch_router, dim_send, dim_send)
        dim_receive = self.dim_receiver(dim_send, dim_buffer, dim_buffer)
        dim_enc = dim_send + self.dropout(dim_receive)
        dim_enc = self.norm3(dim_enc)
        dim_enc = dim_enc + self.dropout(self.MLP2(dim_enc))
        dim_enc = self.norm4(dim_enc)

        final_out = rearrange(dim_enc, '(b seg_num) ts_d d_model -> b ts_d seg_num d_model', b = batch)

        return final_out




import math

import torch
import torch.nn as nn
from torch.nn import functional as F


# -----------------------------------------------------------------------------

class NewGELU(nn.Module):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT).
    Reference: Gaussian Error Linear Units (GELU) paper: https://arxiv.org/abs/1606.08415
    """
    def forward(self, x):
        return 0.5 * x * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))))

class CausalSelfAttention(nn.Module):
    """
    A vanilla multi-head masked self-attention layer with a projection at the end.
    It is possible to use torch.nn.MultiheadAttention here but I am including an
    explicit implementation here to show that there is nothing too scary here.
    """

    def __init__(self, n_embd,n_head,attn_pdrop,resid_pdrop,block_size):
        super().__init__()
        assert n_embd % n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(n_embd, 3 * n_embd)
        # output projection
        self.c_proj = nn.Linear(n_embd, n_embd)
        # regularization
        self.attn_dropout = nn.Dropout(attn_pdrop)
        self.resid_dropout = nn.Dropout(resid_pdrop)
        # causal mask to ensure that attention is only applied to the left in the input sequence
        self.register_buffer("bias", torch.tril(torch.ones(block_size, block_size))
                                     .view(1, 1, block_size, block_size))
        self.n_head = n_head
        self.n_embd = n_embd

    def forward(self, x):
        B, T, C = x.size() # batch size, sequence length, embedding dimensionality (n_embd)

        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        q, k ,v  = self.c_attn(x).split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)

        # causal self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
        att = att.masked_fill(self.bias[:,:,:T,:T] == 0, float('-inf'))
        att = F.softmax(att, dim=-1)
        att = self.attn_dropout(att)
        y = att @ v # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        y = y.transpose(1, 2).contiguous().view(B, T, C) # re-assemble all head outputs side by side

        # output projection
        y = self.resid_dropout(self.c_proj(y))
        return y

class Block(nn.Module):
    """ an unassuming Transformer block """

    def __init__(self, n_embd,resid_pdrop,n_head,attn_pdrop,block_size):
        super().__init__()
        self.ln_1 = nn.LayerNorm(n_embd)
        self.attn = CausalSelfAttention(n_embd,n_head,attn_pdrop,resid_pdrop,block_size)
        self.ln_2 = nn.LayerNorm(n_embd)
        self.mlp = nn.ModuleDict(dict(
            c_fc    = nn.Linear(n_embd, 4 * n_embd),
            c_proj  = nn.Linear(4 * n_embd, n_embd),
            act     = NewGELU(),
            dropout = nn.Dropout(resid_pdrop),
        ))
        m = self.mlp
        self.mlpf = lambda x: m.dropout(m.c_proj(m.act(m.c_fc(x)))) # MLP forward

    def forward(self, x):
        x = x + self.attn(self.ln_1(x))
        x = x + self.mlpf(self.ln_2(x))
        return x




## Copyright 2022 DLinear Authors (https://github.com/cure-lab/LTSF-Linear/tree/main?tab=Apache-2.0-1-ov-file#readme)
## Code modified for align the notation and the batch generation
## extended to all present in informer, autoformer folder

from torch import nn
import torch

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from .utils import QuantileLossMO, get_activation
from typing import List, Union
from ..data_structure.utils import beauty_string
from .utils import  get_scope
from .utils import Embedding_cat_variables
    



class Simple(Base):
    handle_multivariate = True
    handle_future_covariates = True
    handle_categorical_variables = True
    handle_quantile_loss = True
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    description+='\n THE SIMPLE IMPLEMENTATION DOES NOT USE CATEGORICAL NOR FUTURE VARIABLES'
    
    def __init__(self, 
               
                 hidden_size:int,
                 dropout_rate:float=0.1,
                 activation:str='torch.nn.ReLU',


                 **kwargs)->None:
       
        super().__init__(**kwargs)

        if activation == 'torch.nn.SELU':
            beauty_string('SELU do not require BN','info',self.verbose)
            use_bn = False
            
        if isinstance(activation, str):
            activation = get_activation(activation)
        else:
            beauty_string('There is a bug in pytorch lightening, the constructior is called twice','info',self.verbose)
        
        self.save_hyperparameters(logger=False)
      
        
       
        self.emb_past = Embedding_cat_variables(self.past_steps,self.emb_dim,self.embs_past, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        self.emb_fut = Embedding_cat_variables(self.future_steps,self.emb_dim,self.embs_fut, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        emb_past_out_channel = self.emb_past.output_channels
        emb_fut_out_channel = self.emb_fut.output_channels





        self.linear = (nn.Sequential(nn.Linear(emb_past_out_channel*self.past_steps+emb_fut_out_channel*self.future_steps+self.past_steps*self.past_channels+self.future_channels*self.future_steps,hidden_size),
                                                    activation(),nn.Dropout(dropout_rate),
                                                    nn.Linear(hidden_size,self.out_channels*self.future_steps*self.mul)))
    def can_be_compiled(self):
        return True  
                             
    def forward(self, batch):
      
        x =  batch['x_num_past'].to(self.device)
        
        BS = x.shape[0]
        if 'x_cat_future' in batch.keys():
            emb_fut = self.emb_fut(BS,batch['x_cat_future'].to(self.device))
        else:
            emb_fut = self.emb_fut(BS,None)
        if 'x_cat_past' in batch.keys():
            emb_past = self.emb_past(BS,batch['x_cat_past'].to(self.device))
        else:
            emb_past = self.emb_past(BS,None)
        
        
            

        if 'x_num_future' in batch.keys():
            x_future = batch['x_num_future'].to(self.device)
        else:
            x_future = None
            
        tmp = [x,emb_past]
        tot_past = torch.cat(tmp,2).flatten(1)
        


        tmp = [emb_fut]
                          
        if x_future is not None:
            tmp.append(x_future)
           
        tot_future = torch.cat(tmp,2).flatten(1)
        tot = torch.cat([tot_past,tot_future],1)

                    
        res = self.linear(tot)
        res = res.reshape(BS,self.future_steps,-1,self.mul)
       
        ##

        return res
    


import torch.nn as nn
import torch.nn.functional as F

class DecoderLayer(nn.Module):
    def __init__(self, self_attention, cross_attention, d_model, d_ff=None,
                 dropout=0.1, activation="relu"):
        super(DecoderLayer, self).__init__()
        d_ff = d_ff or 4*d_model
        self.self_attention = self_attention
        self.cross_attention = cross_attention
        self.conv1 = nn.Conv1d(in_channels=d_model, out_channels=d_ff, kernel_size=1)
        self.conv2 = nn.Conv1d(in_channels=d_ff, out_channels=d_model, kernel_size=1)
        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.norm3 = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)
        self.activation = F.relu if activation == "relu" else F.gelu

    def forward(self, x, cross, x_mask=None, cross_mask=None):
        x = x + self.dropout(self.self_attention(
            x, x, x,
            attn_mask=x_mask
        )[0])
        x = self.norm1(x)

        x = x + self.dropout(self.cross_attention(
            x, cross, cross,
            attn_mask=cross_mask
        )[0])

        y = x = self.norm2(x)
        y = self.dropout(self.activation(self.conv1(y.transpose(-1,1))))
        y = self.dropout(self.conv2(y).transpose(-1,1))

        return self.norm3(x+y)

class Decoder(nn.Module):
    def __init__(self, layers, norm_layer=None):
        super(Decoder, self).__init__()
        self.layers = nn.ModuleList(layers)
        self.norm = norm_layer

    def forward(self, x, cross, x_mask=None, cross_mask=None):
        for layer in self.layers:
            x = layer(x, cross, x_mask=x_mask, cross_mask=cross_mask)

        if self.norm is not None:
            x = self.norm(x)

        return x



## Copyright Copyright (c) 2020 Andrej Karpathy https://github.com/karpathy/minGPT?tab=MIT-1-ov-file#readme
## Modified for notation alignmenet,  batch structure and adapted for timeseries
## extended to what inside vva folder

from torch import  nn
import torch
from torch.nn import functional as F

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from typing import List, Union
from .vva.minigpt import Block
from .vva.vqvae import VQVAE
import logging
from random import random
from ..data_structure.utils import beauty_string
from .utils import  get_scope

torch.autograd.set_detect_anomaly(True)

        

class VQVAEA(Base):
    handle_multivariate = False
    handle_future_covariates = False
    handle_categorical_variables = False
    handle_quantile_loss = False
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    
    def __init__(self, 
                 past_steps:int,
                 future_steps:int,
                 past_channels:int,
                 future_channels:int,
                 hidden_channels:int,
                 embs:List[int],
                 d_model:int,
                 max_voc_size:int,
                 num_layers:int,
                 dropout_rate:float,
                 commitment_cost:float,
                 decay:float,
                 n_heads:int,
                 out_channels:int,
                 epoch_vqvae: int,
                 persistence_weight:float=0.0,
                 loss_type: str='l1',
                 quantiles:List[int]=[],
                 optim:Union[str,None]=None,
                 optim_config:dict=None,
                 scheduler_config:dict=None,
                 **kwargs)->None:
        """ Custom encoder-decoder 
        
        Args:
            past_steps (int):  number of past datapoints used 
            future_steps (int): number of future lag to predict
            past_channels (int): number of numeric past variables, must be >0
            future_channels (int): number of future numeric variables 
            embs (List): list of the initial dimension of the categorical variables
            cat_emb_dim (int): final dimension of each categorical variable
            hidden_RNN (int): hidden size of the RNN block
            num_layers_RNN (int): number of RNN layers
            kind (str): one among GRU or LSTM
            kernel_size (int): kernel size in the encoder convolutional block
            sum_emb (bool): if true the contribution of each embedding will be summed-up otherwise stacked
            out_channels (int):  number of output channels
            activation (str, optional): activation fuction function pytorch. Default torch.nn.ReLU
            remove_last (bool, optional): if True the model learns the difference respect to the last seen point
            persistence_weight (float):  weight controlling the divergence from persistence model. Default 0
            loss_type (str, optional): this model uses custom losses or l1 or mse. Custom losses can be linear_penalization or exponential_penalization. Default l1,
            quantiles (List[int], optional): we can use quantile loss il len(quantiles) = 0 (usually 0.1,0.5, 0.9) or L1loss in case len(quantiles)==0. Defaults to [].
            dropout_rate (float, optional): dropout rate in Dropout layers
            use_bn (bool, optional): if true BN layers will be added and dropouts will be removed
            use_glu (bool,optional): use GLU for feature selection. Defaults to True.
            glu_percentage (float, optiona): percentage of features to use. Defaults to 1.0.
            n_classes (int): number of classes (0 in regression)
            optim (str, optional): if not None it expects a pytorch optim method. Defaults to None that is mapped to Adam.
            optim_config (dict, optional): configuration for Adam optimizer. Defaults to None.
            scheduler_config (dict, optional): configuration for stepLR scheduler. Defaults to None.

        """

        
        super().__init__(**kwargs)
        self.save_hyperparameters(logger=False)
        self.d_model = d_model
        self.max_voc_size = max_voc_size
        self.future_steps = future_steps
        self.epoch_vqvae = epoch_vqvae
        ##PRIMA VQVAE
        assert out_channels==1, beauty_string('Working only for one singal','section',True)
        assert past_steps%2==0 and future_steps%2==0, beauty_string('There are some issue with the deconder in case of odd length','section',True)
        self.vqvae = VQVAE(in_channels=1, hidden_channels=hidden_channels,out_channels=1,num_embeddings= max_voc_size,embedding_dim=d_model,commitment_cost=commitment_cost,decay=decay  )
        
        ##POI GPT


        self.block_size = past_steps//2 + future_steps//2 -1
        self.sentence_length = future_steps//2
        
        self.transformer = nn.ModuleDict(dict(
            wte = nn.Embedding(max_voc_size, d_model),
            wpe = nn.Embedding(self.block_size, d_model),
            drop = nn.Dropout(dropout_rate),
            h = nn.ModuleList([Block( d_model,dropout_rate,n_heads,dropout_rate,self.block_size) for _ in range(num_layers)]), ##care can be different dropouts
            ln_f = nn.LayerNorm(d_model),
            lm_head = nn.Linear(d_model, max_voc_size, bias=False)
        ))
        # report number of parameters (note we don't count the decoder parameters in lm_head)
        n_params = sum(p.numel() for p in self.transformer.parameters())
        beauty_string("number of parameters: %.2fM" % (n_params/1e6,),'info',self.verbose)
        
        self.use_quantiles = False
        self.is_classification = True
        self.optim_config = optim_config
       
        

        
    def configure_optimizers(self):
        
    
        #return torch.optim.Adam(self.vqvae.parameters(), lr=self.optim_config.lr_vqvae,
        #                                   weight_decay=self.optim_config.weight_decay_vqvae)


        return torch.optim.AdamW([
                {'params':self.vqvae.parameters(),'lr':self.optim_config.lr_vqvae,'weight_decay':self.optim_config.weight_decay_vqvae},
                {'params':self.transformer.parameters(),'lr':self.optim_config.lr_gpt,'weight_decay':self.optim_config.weight_decay_gpt},
            ])
       

        
    def gpt(self,tokens):
    
    
        b, t = tokens.size()
        assert t <= self.block_size, beauty_string("Cannot forward sequence of length {t}, block size is only {self.block_size}",'section',True)
        pos = torch.arange(0, t, dtype=torch.long, device=self.device).unsqueeze(0) # shape (1, t)

        # forward the GPT model itself
        tok_emb = self.transformer.wte(tokens) # token embeddings of shape (b, t, n_embd)
        pos_emb = self.transformer.wpe(pos) # position embeddings of shape (1, t, n_embd)
        x = self.transformer.drop(tok_emb + pos_emb)
        for block in self.transformer.h:
            x = block(x)
        x = self.transformer.ln_f(x)
        logits = self.transformer.lm_head(x)
        return logits

    def forward(self, batch):
        
        ##VQVAE
        #current_epoch = self.current_epoch 
        #if current_epoch < 1000:
        #    self.vqvae.train()
        #    loss_gpt = 100
        #else:
        #    self.vqvae.eval()
        idx_target = batch['idx_target'][0]

        #(tensor([194, 163, 174, 176, 160, 168, 175]),
        # tensor([ -1,  -1,  -1, 160, 168, 175, 160]))

           
        data = batch['x_num_past'][:,:,idx_target]
        if self.current_epoch > self.epoch_vqvae:
            with torch.no_grad():
                vqloss, data_recon, perplexity,quantized_x,encodings_x = self.vqvae(data.permute(0,2,1))
                loss_vqvae = 0
        else:
            vq_loss, data_recon, perplexity,quantized_x,encodings_x = self.vqvae(data.permute(0,2,1))
            if random()<0.001:
                beauty_string(perplexity,'info',self.verbose)
            recon_error = F.mse_loss(data_recon.squeeze(), data.squeeze()) 
            loss_vqvae = recon_error + vq_loss

        if self.current_epoch > self.epoch_vqvae:
            with torch.no_grad():
                _, _, _,quantized_y,encodings_y = self.vqvae(batch['y'].permute(0,2,1))
            
            ##GPT
            tokens = torch.cat([encodings_x.argmax(dim=2),encodings_y.argmax(dim=2)[:,0:-1]],1)
            tokens_y = torch.cat([encodings_x.argmax(dim=2)[:,0:-1],encodings_y.argmax(dim=2)],1)
            tokens_y[:,0:encodings_x.shape[1]-1] = -1
            logits = self.gpt(tokens)
            loss_gpt = F.cross_entropy(logits.view(-1, logits.size(-1)),tokens_y.view(-1), ignore_index=-1)

            
            ##adesso devo ricostruire la y perche' e quello che voglio come output
            with torch.no_grad():
                encoding_indices = torch.argmax(logits.reshape(-1,self.max_voc_size), dim=1).unsqueeze(1) ##
                encodings = torch.zeros(encoding_indices.shape[0], self.vqvae._vq_vae._num_embeddings, device=self.device)
                encodings.scatter_(1, encoding_indices, 1)
                quantized = torch.matmul(encodings, self.vqvae._vq_vae._embedding.weight).view(data.shape[0],-1,self.d_model) ##B x L x hidden
                quantized = quantized.permute(0, 2, 1).contiguous()
                y_hat = self.vqvae._decoder(quantized,False).squeeze()[:,-self.future_steps:]
            
            l1_loss = nn.L1Loss()(y_hat,batch['y'].squeeze())

            return y_hat, loss_vqvae+loss_gpt+l1_loss
        else:
            return None, loss_vqvae


    def training_step(self, batch, batch_idx):
        """
        pythotrch lightening stuff
        
        :meta private:
        """
        _, loss = self(batch)
        return loss
    
    def validation_step(self, batch, batch_idx):
        """
        pythotrch lightening stuff
        
        :meta private:
        """
        _, loss = self(batch)
        return loss


    def generate(self, idx, max_new_tokens, temperature=1.0, do_sample=False, top_k=None,num_samples=100):
        """
        Take a conditioning sequence of indices idx (LongTensor of shape (b,t)) and complete
        the sequence max_new_tokens times, feeding the predictions back into the model each time.
        Most likely you'll want to make sure to be in model.eval() mode of operation for this.
        """
        assert do_sample is False,logging.info('NOT IMPLEMENTED YET')
        if do_sample:
            
            idx = idx.repeat(num_samples,1,1)
            for _ in range(max_new_tokens):
                tmp = []
                for i in range(num_samples):
                    idx_cond = idx[i,:,:] if idx.size(2) <= self.block_size else idx[i,:, -self.block_size:]
                    logits = self.gpt(idx_cond)
                    logits = logits[:, -1, :] / temperature
                    if top_k is not None:
                        v, _ = torch.topk(logits, top_k)
                        logits[logits < v[:, [-1]]] = -float('Inf')
                    probs = F.softmax(logits, dim=-1)
                    idx_next = torch.multinomial(probs, num_samples=1, replacement=True)
                    tmp.append(idx_next)
                tmp = torch.cat(tmp,dim=1).T.unsqueeze(2)
                idx = torch.cat((idx, tmp), dim=2)
            return idx
        else:
            for _ in range(max_new_tokens):
                
                # if the sequence context is growing too long we must crop it at block_size
                idx_cond = idx if idx.size(1) <= self.block_size else idx[:, -self.block_size:]
                # forward the model to get the logits for the index in the sequence
                logits = self.gpt(idx_cond)
                # pluck the logits at the final step and scale by desired temperature
                logits = logits[:, -1, :] / temperature
                # optionally crop the logits to only the top k options
                if top_k is not None:
                    v, _ = torch.topk(logits, top_k)
                    logits[logits < v[:, [-1]]] = -float('Inf')
                # apply softmax to convert logits to (normalized) probabilities
                probs = F.softmax(logits, dim=-1)
                # either sample from the distribution or take the most likely element
                _, idx_next = torch.topk(probs, k=1, dim=-1)
                # append sampled index to the running sequence and continue
                idx = torch.cat((idx, idx_next), dim=1)

            return idx

    def inference(self, batch:dict)->torch.tensor:

        idx_target = batch['idx_target'][0]
        data = batch['x_num_past'][:,:,idx_target].to(self.device)
        vq_loss, data_recon, perplexity,quantized_x,encodings_x = self.vqvae(data.permute(0,2,1))
        x = encodings_x.argmax(dim=2)
        inp = x[:, :self.sentence_length]
        # let the model sample the rest of the sequence
        cat = self.generate(inp, self.sentence_length, do_sample=False) # non riesco a gestirla qui :-)
        encoding_indices = cat.flatten().unsqueeze(1) ##
        encodings = torch.zeros(encoding_indices.shape[0], self.vqvae._vq_vae._num_embeddings, device=self.device)
        encodings.scatter_(1, encoding_indices, 1)
        quantized = torch.matmul(encodings, self.vqvae._vq_vae._embedding.weight).view(x.shape[0],-1,self.d_model) ##B x L x hidden
        quantized = quantized.permute(0, 2, 1).contiguous()
        y_hat = self.vqvae._decoder(quantized,False).squeeze()[:,-self.future_steps:]

        ## BxLxCx3
        return y_hat.unsqueeze(2).unsqueeze(3)
    



import torch
import torch.nn as nn
from .tft import sub_nn

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from .utils import QuantileLossMO
from typing import List, Union
from ..data_structure.utils import beauty_string
from .utils import  get_scope
from .utils import Embedding_cat_variables

class TFT(Base):
    handle_multivariate = True
    handle_future_covariates = True
    handle_categorical_variables = True
    handle_quantile_loss = True
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    def __init__(self, 
                 d_model: int,
                 num_layers_RNN: int,
                 d_head: int,
                 n_head: int,
                 dropout_rate: float,
                
                 **kwargs)->None:
        """Initializes the model for time series forecasting with attention mechanisms and recurrent neural networks.
        
        This model is designed for direct forecasting, allowing for multi-output and multi-horizon predictions. It leverages attention mechanisms to enhance the selection of relevant past time steps and learn long-term dependencies. The architecture includes RNN enrichment, gating mechanisms to minimize the impact of irrelevant variables, and the ability to output prediction intervals through quantile regression.
        
        Key features include:
        - Direct Model: Predicts all future steps at once.
        - Multi-Output Forecasting: Capable of predicting one or more variables simultaneously.
        - Multi-Horizon Forecasting: Predicts variables at multiple future time steps.
        - Attention-Based Mechanism: Enhances the selection of relevant past time steps and learns long-term dependencies.
        - RNN Enrichment: Utilizes LSTM for initial autoregressive approximation, which is refined by the rest of the network.
        - Gating Mechanisms: Reduces the contribution of irrelevant variables.
        - Prediction Intervals: Outputs percentiles (e.g., 10th, 50th, 90th) at each time step.
        
        The model also facilitates interpretability by identifying:
        - Global importance of variables for both past and future.
        - Temporal patterns.
        - Significant events.
        
        Args:
            d_model (int): General hidden dimension across the network, adjustable in sub-networks.
            num_layers_RNN (int): Number of layers in the recurrent neural network (LSTM).
            d_head (int): Dimension of each attention head.
            n_head (int): Number of attention heads.
            dropout_rate (float): Dropout rate applied uniformly across all dropout layers.
            **kwargs: Additional keyword arguments for further customization.
        """
        
        
        


        super().__init__(**kwargs)
        self.save_hyperparameters(logger=False)
        # assert out_channels==1, logging.info("ONLY ONE CHANNEL IMPLEMENTED")
        self.d_model = d_model
        # linear to embed the target vartiable
        self.target_linear = nn.Linear(self.out_channels, d_model) # same for past and fut! (same variable)
        # number of variables in the past different from the target one(s)
        self.aux_past_channels = self.past_channels - self.out_channels # -1 because one channel is occupied by the target variable
        # one linear for each auxiliar past var
        self.linear_aux_past = nn.ModuleList([nn.Linear(1, d_model) for _ in range(self.aux_past_channels)])
        # number of variables in the future used to predict the target one(s)
        self.aux_fut_channels = self.future_channels
        # one linear for each auxiliar future var
        self.linear_aux_fut = nn.ModuleList([nn.Linear(1, d_model) for _ in range(self.aux_fut_channels)])
        # length of the full sequence, parameter used for the embedding of all categorical variables
        # - we assume that these are no available or available both for past and future
        seq_len = self.past_steps+self.future_steps
        
        
        ##in v.1.1.5 this is not working, past and future are different for categorical
        #self.emb_cat_var = sub_nn.embedding_cat_variables(seq_len, self.future_steps, d_model, embs, self.device)
        
        self.emb_past = Embedding_cat_variables(self.past_steps,self.emb_dim,self.embs_past, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        self.emb_fut = Embedding_cat_variables(self.future_steps,self.emb_dim,self.embs_fut, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        emb_past_out_channel = self.emb_past.output_channels
        emb_fut_out_channel = self.emb_fut.output_channels
        
        
        # Recurrent Neural Network for first aproximated inference of the target variable(s) - IT IS NON RE-EMBEDDED YET
        self.rnn = sub_nn.LSTM_Model(num_var=self.out_channels, 
                                     d_model = d_model, 
                                     pred_step = self.future_steps, 
                                     num_layers = num_layers_RNN, 
                                     dropout = dropout_rate)
        # PARTS OF TFT:
        # - Residual connections
        # - Gated Residual Network
        # - Interpretable MultiHead Attention
        self.res_conn1_past = sub_nn.ResidualConnection(d_model, dropout_rate)
        self.res_conn1_fut = sub_nn.ResidualConnection(d_model, dropout_rate)
        self.grn1_past = sub_nn.GRN(d_model, dropout_rate)
        self.grn1_fut = sub_nn.GRN(d_model, dropout_rate)
        self.InterpretableMultiHead = sub_nn.InterpretableMultiHead(d_model, d_head, n_head)
        self.res_conn2_att = sub_nn.ResidualConnection(d_model, dropout_rate)
        self.grn2_att = sub_nn.GRN(d_model, dropout_rate)
        self.res_conn3_out = sub_nn.ResidualConnection(d_model, dropout_rate)

        self.outLinear = nn.Linear(d_model, self.out_channels*self.mul)

    def can_be_compiled(self):
        return False  
  

    def forward(self, batch:dict) -> torch.Tensor:
        """Temporal Fusion Transformer

        Collectiong Data
        - Extract the autoregressive variable(s)
        - Embedding and compute a first approximated prediction
        - 'summary_past' and 'summary_fut' collecting data about past and future
        Concatenating on the dimension 2 all different datas, which will be mixed through a MEAN over that imension
        Info get from other tensor of the batch taken as input
        
        TFT actual computations
        - Residual Connection for y_past and summary_past
        - Residual Connection for y_fut and summary_fut
        - GRN1 for past and for fut
        - ATTENTION(summary_fut, summary_past, y_past) 
        - Residual Connection for attention itself
        - GRN2 for attention
        - Residual Connection for attention and summary_fut
        - Linear for actual values and reshape 

        Args:
            batch (dict): Keys used are ['x_num_past', 'idx_target', 'x_num_future', 'x_cat_past', 'x_cat_future']

        Returns:
            torch.Tensor: shape [B, self.future_steps, self.out_channels, self.mul] or [B, self.future_steps, self.out_channels] according to quantiles
        """

        num_past = batch['x_num_past'].to(self.device)
        # PAST TARGET NUMERICAL VARIABLE
        # always available: autoregressive variable
        # compute rnn prediction
        idx_target = batch['idx_target'][0]
        target_num_past = num_past[:,:,idx_target]
        target_emb_num_past = self.target_linear(target_num_past) # target_variables comunicating with each others
        target_num_fut_approx = self.rnn(target_emb_num_past)
        # embed future predictions
        target_emb_num_fut_approx = self.target_linear(target_num_fut_approx)

        ### create variable summary_past and summary_fut
        # at the beggining it is composed only by past and future target variable
        summary_past = target_emb_num_past.unsqueeze(2)
        summary_fut = target_emb_num_fut_approx.unsqueeze(2)
        # now we search for others categorical and numerical variables!


        ### PAST NUMERICAL VARIABLES
        if self.aux_past_channels>0: # so we have more numerical variables about past
            # AUX = AUXILIARY variables
            aux_num_past = self.remove_var(num_past, idx_target, 2) # remove the target index on the second dimension
            assert self.aux_past_channels == aux_num_past.size(2), beauty_string(f"{self.aux_past_channels} LAYERS FOR PAST VARS AND {aux_num_past.shape(2)} VARS",'section',True) # to check if we are using the expected number of variables about past
            aux_emb_num_past = torch.Tensor().to(aux_num_past.device)
            for i, layer in enumerate(self.linear_aux_past):
                aux_emb_past = layer(aux_num_past[:,:,[i]]).unsqueeze(2)
                aux_emb_num_past = torch.cat((aux_emb_num_past, aux_emb_past), dim=2)
            ## update summary about past
            summary_past = torch.cat((summary_past, aux_emb_num_past), dim=2)
        
        ### FUTURE NUMERICAL VARIABLES
        if self.aux_fut_channels>0: # so we have more numerical variables about future
            aux_num_fut = batch['x_num_future'].to(self.device)
            assert self.aux_fut_channels == aux_num_fut.size(2), beauty_string(f"{self.aux_fut_channels} LAYERS FOR PAST VARS AND {aux_num_fut.size(2)} VARS",'section',True)  # to check if we are using the expected number of variables about fut
            aux_emb_num_fut = torch.Tensor().to(aux_num_fut.device)
            for j, layer in enumerate(self.linear_aux_fut):
                aux_emb_fut = layer(aux_num_fut[:,:,[j]]).unsqueeze(2)
                aux_emb_num_fut = torch.cat((aux_emb_num_fut, aux_emb_fut), dim=2)
            ## update summary about future
            summary_fut = torch.cat((summary_fut, aux_emb_num_fut), dim=2)
        '''
        ### CATEGORICAL VARIABLES changed in 1.1.5
        if 'x_cat_past' in batch.keys() and 'x_cat_future' in batch.keys(): # if we have both
            # HERE WE ASSUME SAME NUMBER AND KIND OF VARIABLES IN PAST AND FUTURE
            cat_past = batch['x_cat_past'].to(self.device)
            cat_fut = batch['x_cat_future'].to(self.device)
            cat_full = torch.cat((cat_past, cat_fut), dim = 1)
            # EMB CATEGORICAL VARIABLES AND THEN SPLIT IN PAST AND FUTURE
            emb_cat_full = self.emb_cat_var(cat_full,self.device)
        else:
            emb_cat_full = self.emb_cat_var(num_past.shape[0],self.device)
            
        cat_emb_past = emb_cat_full[:,:-self.future_steps,:,:]
        cat_emb_fut = emb_cat_full[:,-self.future_steps:,:,:]
        
        ## update summary
        # past
        summary_past = torch.cat((summary_past, cat_emb_past), dim=2)
        # future
        summary_fut = torch.cat((summary_fut, cat_emb_fut), dim=2)
        '''
        BS = num_past.shape[0]
        if 'x_cat_future' in batch.keys():
            emb_fut = self.emb_fut(BS,batch['x_cat_future'].to(self.device))
        else:
            emb_fut = self.emb_fut(BS,None)
        if 'x_cat_past' in batch.keys():
            emb_past = self.emb_past(BS,batch['x_cat_past'].to(self.device))
        else:
            emb_past = self.emb_past(BS,None)
            
        ## update summary
        # past

        summary_past = torch.cat((summary_past, emb_past.unsqueeze(-1).repeat((1,1,1,summary_past.shape[-1]))), dim=2)
        # future
        summary_fut = torch.cat((summary_fut, emb_fut.unsqueeze(-1).repeat((1,1,1,summary_past.shape[-1]))), dim=2)

        # >>> PAST:
        summary_past = torch.mean(summary_past, dim=2)
        # >>> FUTURE:
        summary_fut = torch.mean(summary_fut, dim=2)

        ### Residual Connection from LSTM
        summary_past = self.res_conn1_past(summary_past, target_emb_num_past)
        summary_fut = self.res_conn1_fut(summary_fut, target_emb_num_fut_approx)

        ### GRN1
        summary_past = self.grn1_past(summary_past)
        summary_fut = self.grn1_fut(summary_fut)

        ### INTERPRETABLE MULTI HEAD ATTENTION
        attention = self.InterpretableMultiHead(summary_fut, summary_past, target_emb_num_past)

        ### Residual Connection from ATT
        attention = self.res_conn2_att(attention, attention)

        ### GRN
        attention = self.grn2_att(attention)

        ### Resuidual Connection from GRN1
        out = self.res_conn3_out(attention, summary_fut)

        ### OUT
        out = self.outLinear(out)

        if self.mul>0:
            out = out.view(-1, self.future_steps, self.out_channels, self.mul)
        return out
    
    #function to extract from batch['x_num_past'] all variables except the one autoregressive
    def remove_var(self, tensor: torch.Tensor, indexes_to_exclude: int, dimension: int)-> torch.Tensor:
        """Function to remove variables from tensors in chosen dimension and position 

        Args:
            tensor (torch.Tensor): starting tensor
            indexes_to_exclude (int): index of the chosen dimension we want t oexclude
            dimension (int): dimension of the tensor on which we want to work

        Returns:
            torch.Tensor: new tensor without the chosen variables
        """

        remaining_idx = torch.tensor([i for i in range(tensor.size(dimension)) if i not in indexes_to_exclude]).to(tensor.device)
        # Select the desired sub-tensor
        extracted_subtensors = torch.index_select(tensor, dim=dimension, index=remaining_idx)
        
        return extracted_subtensors
    
